# -*- coding: utf-8 -*-
"""Test adding header lines to headers
"""

import io

import pytest

import vcfpy
from vcfpy import parser
from vcfpy import header

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"

HEADER = """
##fileformat=VCFv4.3
##fileDate=20090805
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00001\tNA00002\tNA00003
""".lstrip()


@pytest.fixture(scope="function")
def vcf_header():
    p = parser.Parser(io.StringIO(HEADER))
    return p.parse_header()


def test_add_simple_line(vcf_header):
    assert len(vcf_header.lines) == 18
    line = header.HeaderLine("somekey", "somevalue")
    vcf_header.add_line(line)
    assert len(vcf_header.lines) == 19

    assert vcf_header.lines[-1].key == "somekey"
    assert vcf_header.lines[-1].value == "somevalue"


def test_add_contig_line(vcf_header):
    # check header before adding
    assert len(vcf_header.lines) == 18
    assert "20a" not in vcf_header._indices["contig"]

    # add header line
    VALUE = (
        "<ID=20a,length=62435964,assembly=B36,"
        "md5=f126cdf8a6e0c7f379d618ff66beb2da,"
        'species="Homo sapiens",taxonomy=x>'
    )
    line = header.ContigHeaderLine(
        "contig",
        VALUE,
        vcfpy.OrderedDict(
            [
                ("ID", "20a"),
                ("length", 62435964),
                ("assembly", "B36"),
                ("md5", "f126cdf8a6e0c7f379d618ff66beb2da"),
                ("species", "Homo sapiens"),
                ("taxonomy", "x"),
            ]
        ),
    )
    vcf_header.add_line(line)

    # check header after adding
    assert len(vcf_header.lines) == 19
    assert "20a" in vcf_header._indices["contig"]
    assert vcf_header._indices["contig"]["20a"] is vcf_header.lines[-1]

    # Check resulting added header line
    assert vcf_header.lines[-1].key == "contig"
    assert vcf_header.lines[-1].value == VALUE
    assert len(vcf_header.lines[-1].mapping) == 6
    assert vcf_header.lines[-1].mapping["ID"] == "20a"
    assert vcf_header.lines[-1].mapping["length"] == 62435964
    assert vcf_header.lines[-1].mapping["assembly"] == "B36"
    assert vcf_header.lines[-1].mapping["md5"] == "f126cdf8a6e0c7f379d618ff66beb2da"
    assert vcf_header.lines[-1].mapping["species"] == "Homo sapiens"
    assert vcf_header.lines[-1].mapping["taxonomy"] == "x"


def test_add_contig_line_shortcut(vcf_header):
    # check header before adding
    assert len(vcf_header.lines) == 18
    assert "20a" not in vcf_header._indices["contig"]

    # add header line
    mapping = vcfpy.OrderedDict(
        [
            ("ID", "20a"),
            ("length", 62435964),
            ("assembly", "B36"),
            ("md5", "f126cdf8a6e0c7f379d618ff66beb2da"),
            ("species", "Homo sapiens"),
            ("taxonomy", "x"),
        ]
    )
    vcf_header.add_contig_line(mapping)

    # check header after adding
    assert len(vcf_header.lines) == 19
    assert "20a" in vcf_header._indices["contig"]
    assert vcf_header._indices["contig"]["20a"] is vcf_header.lines[-1]

    # Check resulting added header line
    assert vcf_header.lines[-1].key == "contig"
    VALUE = (
        "<ID=20a,length=62435964,assembly=B36,"
        "md5=f126cdf8a6e0c7f379d618ff66beb2da,"
        'species="Homo sapiens",taxonomy=x>'
    )
    assert vcf_header.lines[-1].value == VALUE
    assert len(vcf_header.lines[-1].mapping) == 6
    assert vcf_header.lines[-1].mapping["ID"] == "20a"
    assert vcf_header.lines[-1].mapping["length"] == 62435964
    assert vcf_header.lines[-1].mapping["assembly"] == "B36"
    assert vcf_header.lines[-1].mapping["md5"] == "f126cdf8a6e0c7f379d618ff66beb2da"
    assert vcf_header.lines[-1].mapping["species"] == "Homo sapiens"
    assert vcf_header.lines[-1].mapping["taxonomy"] == "x"


def test_add_filter_line(vcf_header):
    # check header before adding
    assert len(vcf_header.lines) == 18

    # add header line
    VALUE = '<ID=q10a,Description="Quality below 10">'
    line = header.FilterHeaderLine(
        "FILTER", VALUE, vcfpy.OrderedDict([("ID", "q10a"), ("Description", "Quality below 10")])
    )
    vcf_header.add_line(line)

    # check header after adding
    assert len(vcf_header.lines) == 19
    assert "q10a" in vcf_header._indices["FILTER"]
    assert vcf_header._indices["FILTER"]["q10a"] is vcf_header.lines[-1]

    # Check resulting added header line
    assert vcf_header.lines[-1].key == "FILTER"
    assert vcf_header.lines[-1].value == VALUE
    assert len(vcf_header.lines[-1].mapping) == 2
    assert vcf_header.lines[-1].mapping["ID"] == "q10a"
    assert vcf_header.lines[-1].mapping["Description"] == "Quality below 10"


def test_add_filter_line_shortcut(vcf_header):
    # check header before adding
    assert len(vcf_header.lines) == 18

    # add header line
    mapping = vcfpy.OrderedDict([("ID", "q10a"), ("Description", "Quality below 10")])
    vcf_header.add_filter_line(mapping)

    # check header after adding
    assert len(vcf_header.lines) == 19
    assert "q10a" in vcf_header._indices["FILTER"]
    assert vcf_header._indices["FILTER"]["q10a"] is vcf_header.lines[-1]

    # Check resulting added header line
    assert vcf_header.lines[-1].key == "FILTER"
    VALUE = '<ID=q10a,Description="Quality below 10">'
    assert vcf_header.lines[-1].value == VALUE
    assert len(vcf_header.lines[-1].mapping) == 2
    assert vcf_header.lines[-1].mapping["ID"] == "q10a"
    assert vcf_header.lines[-1].mapping["Description"] == "Quality below 10"


def test_add_format_line(vcf_header):
    # check header before adding
    assert len(vcf_header.lines) == 18

    # add header line
    VALUE = '<ID=GTa,Number=1,Type=String,Description="Genotype">'
    line = header.FormatHeaderLine(
        "FORMAT",
        VALUE,
        vcfpy.OrderedDict(
            [("ID", "GTa"), ("Number", 1), ("Type", "String"), ("Description", "Genotype")]
        ),
    )
    vcf_header.add_line(line)

    # check header after adding
    assert len(vcf_header.lines) == 19
    assert "GTa" in vcf_header._indices["FORMAT"]
    assert vcf_header._indices["FORMAT"]["GTa"] is vcf_header.lines[-1]

    # Check resulting added header line
    assert vcf_header.lines[-1].key == "FORMAT"
    assert vcf_header.lines[-1].value == VALUE
    assert len(vcf_header.lines[-1].mapping) == 4
    assert vcf_header.lines[-1].mapping["ID"] == "GTa"
    assert vcf_header.lines[-1].mapping["Number"] == 1
    assert vcf_header.lines[-1].mapping["Type"] == "String"
    assert vcf_header.lines[-1].mapping["Description"] == "Genotype"


def test_add_format_line_shortcut(vcf_header):
    # check header before adding
    assert len(vcf_header.lines) == 18

    # add header line
    mapping = vcfpy.OrderedDict(
        [("ID", "GTa"), ("Number", 1), ("Type", "String"), ("Description", "Genotype")]
    )
    vcf_header.add_format_line(mapping)

    # check header after adding
    assert len(vcf_header.lines) == 19
    assert "GTa" in vcf_header._indices["FORMAT"]
    assert vcf_header._indices["FORMAT"]["GTa"] is vcf_header.lines[-1]

    # Check resulting added header line
    assert vcf_header.lines[-1].key == "FORMAT"
    VALUE = '<ID=GTa,Number=1,Type=String,Description="Genotype">'
    assert vcf_header.lines[-1].value == VALUE
    assert len(vcf_header.lines[-1].mapping) == 4
    assert vcf_header.lines[-1].mapping["ID"] == "GTa"
    assert vcf_header.lines[-1].mapping["Number"] == 1
    assert vcf_header.lines[-1].mapping["Type"] == "String"
    assert vcf_header.lines[-1].mapping["Description"] == "Genotype"


def test_add_info_line(vcf_header):
    # check header before adding
    assert len(vcf_header.lines) == 18

    # add header line
    VALUE = '<ID=DPa,Number=1,Type=Integer,Description="Total Depth">'
    line = header.FormatHeaderLine(
        "INFO",
        VALUE,
        vcfpy.OrderedDict(
            [("ID", "DPa"), ("Number", 1), ("Type", "Integer"), ("Description", "Total Depth")]
        ),
    )
    vcf_header.add_line(line)
    assert len(vcf_header.lines) == 19

    # check header after adding
    assert len(vcf_header.lines) == 19
    assert "DPa" in vcf_header._indices["INFO"]
    assert vcf_header._indices["INFO"]["DPa"] is vcf_header.lines[-1]

    # Check resulting added header line
    assert vcf_header.lines[-1].key == "INFO"
    assert vcf_header.lines[-1].value == VALUE
    assert len(vcf_header.lines[-1].mapping) == 4
    assert vcf_header.lines[-1].mapping["ID"] == "DPa"
    assert vcf_header.lines[-1].mapping["Number"] == 1
    assert vcf_header.lines[-1].mapping["Type"] == "Integer"
    assert vcf_header.lines[-1].mapping["Description"] == "Total Depth"


def test_add_info_line_shortcut(vcf_header):
    # check header before adding
    assert len(vcf_header.lines) == 18

    # add header line
    VALUE = '<ID=DPa,Number=1,Type=Integer,Description="Total Depth">'
    mapping = vcfpy.OrderedDict(
        [("ID", "DPa"), ("Number", 1), ("Type", "Integer"), ("Description", "Total Depth")]
    )
    vcf_header.add_info_line(mapping)
    assert len(vcf_header.lines) == 19

    # check header after adding
    assert len(vcf_header.lines) == 19
    assert "DPa" in vcf_header._indices["INFO"]
    assert vcf_header._indices["INFO"]["DPa"] is vcf_header.lines[-1]

    # Check resulting added header line
    assert vcf_header.lines[-1].key == "INFO"
    assert vcf_header.lines[-1].value == VALUE
    assert len(vcf_header.lines[-1].mapping) == 4
    assert vcf_header.lines[-1].mapping["ID"] == "DPa"
    assert vcf_header.lines[-1].mapping["Number"] == 1
    assert vcf_header.lines[-1].mapping["Type"] == "Integer"
    assert vcf_header.lines[-1].mapping["Description"] == "Total Depth"



# -*- coding: utf-8 -*-
"""Test procesing of ALT column
"""

import pytest

from vcfpy import header
from vcfpy import parser

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"

# TODO: add more tests


@pytest.fixture
def vcf_header():
    return header.Header()


# result type Substitution ----------------------------------------------------


def test_substitution_snv(vcf_header):
    EXPECTED = "Substitution(type_='SNV', value='T')"
    RESULT = parser.process_alt(vcf_header, "C", "T")
    assert str(RESULT) == EXPECTED


def test_substitution_mnv(vcf_header):
    EXPECTED = "Substitution(type_='MNV', value='TG')"
    RESULT = parser.process_alt(vcf_header, "GC", "TG")
    assert str(RESULT) == EXPECTED


def test_substitution_ins(vcf_header):
    EXPECTED = "Substitution(type_='INS', value='CT')"
    RESULT = parser.process_alt(vcf_header, "C", "CT")
    assert str(RESULT) == EXPECTED


def test_substitution_del(vcf_header):
    EXPECTED = "Substitution(type_='DEL', value='C')"
    RESULT = parser.process_alt(vcf_header, "CT", "C")
    assert str(RESULT) == EXPECTED


def test_substitution_indel(vcf_header):
    EXPECTED = "Substitution(type_='INDEL', value='TG')"
    RESULT = parser.process_alt(vcf_header, "AAAC", "TG")
    assert str(RESULT) == EXPECTED



# -*- coding: utf-8 -*-
"""Reading of VCF files from plain and bgzip-ed files
"""

import io
import os

from vcfpy import reader

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


def test_read_text():
    path = os.path.join(os.path.dirname(__file__), "vcfs/full_vcf43.vcf")
    r = reader.Reader.from_path(path)
    assert r.parser
    assert r.header
    assert len(r.header.lines) == 18
    assert r.header.samples
    assert r.header.samples.names == ["NA00001", "NA00002", "NA00003"]
    records = []
    for record in r:
        records.append(record)
    assert len(records) == 5


def test_read_bgzip():
    path = os.path.join(os.path.dirname(__file__), "vcfs/full_vcf43.vcf.gz")
    r = reader.Reader.from_path(path)
    assert r.parser
    assert r.header
    assert len(r.header.lines) == 18
    assert r.header.samples
    assert r.header.samples.names == ["NA00001", "NA00002", "NA00003"]
    records = []
    for record in r:
        records.append(record)
    assert len(records) == 5


def test_read_text_no_samples():
    path = os.path.join(os.path.dirname(__file__), "vcfs/full_vcf43_no_samples.vcf")
    r = reader.Reader.from_path(path)
    assert r.parser
    assert r.header
    assert len(r.header.lines) == 18
    assert r.header.samples
    assert r.header.samples.names == []
    records = []
    for record in r:
        records.append(record)
    assert len(records) == 5


def test_read_info_flag():
    """Test reading INFO field flag with inconsistent header metadata."""
    # In the INFO field, `MH` is a flag but the header specifies it as a string.
    string_buffer = io.StringIO(
        r"""##fileformat=VCFv4.2
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##INFO=<ID=MH,Number=1,Type=String,Description="Microhomology">
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	FOO	BAR
1	4798729	.	CT	C	621	PASS	MH	GT	0/0:25	0/1:64
"""
    )

    # Verify that no TypeError is raised.
    with reader.Reader.from_stream(string_buffer) as r:
        line = next(r)
        assert line.INFO["MH"]



# -*- coding: utf-8 -*-

from .compat import OrderedDict

from .exceptions import (
    VCFPyException,
    InvalidHeaderException,
    InvalidRecordException,
    IncorrectVCFFormat,
    HeaderNotFound,
)

from .header import (
    Header,
    HeaderLine,
    SimpleHeaderLine,
    AltAlleleHeaderLine,
    ContigHeaderLine,
    FilterHeaderLine,
    MetaHeaderLine,
    PedigreeHeaderLine,
    SampleHeaderLine,
    CompoundHeaderLine,
    InfoHeaderLine,
    FormatHeaderLine,
    SamplesInfos,
    FieldInfo,
    header_without_lines,
)

from .record import (
    Record,
    Call,
    UnparsedCall,
    AltRecord,
    Substitution,
    BreakEnd,
    SingleBreakEnd,
    SymbolicAllele,
)

from .record import SNV, MNV, DEL, INS, INDEL, SV, BND, SYMBOLIC, MIXED
from .record import HOM_REF, HET, HOM_ALT
from .record import FIVE_PRIME, THREE_PRIME, FORWARD, REVERSE

from .reader import Reader

from .writer import Writer

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



# -*- coding: utf-8 -*-
"""Test fetching within tabix files
"""

import os

from vcfpy import reader

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


# Test fetch with chrom/begin/end ---------------------------------------------


def test_fetch_no_records_values():
    path = os.path.join(os.path.dirname(__file__), "vcfs", "multi_contig.vcf.gz")
    r = reader.Reader.from_path(path)

    records = [vcf_rec for vcf_rec in r.fetch("20", 1110698, 1230236)]

    assert len(records) == 0


def test_fetch_one_record_values():
    path = os.path.join(os.path.dirname(__file__), "vcfs", "multi_contig.vcf.gz")
    r = reader.Reader.from_path(path)

    records = [vcf_rec for vcf_rec in r.fetch("20", 1110695, 1230236)]

    assert len(records) == 1

    assert records[0].CHROM == "20"
    assert records[0].POS == 1110696


def test_fetch_two_records_values():
    path = os.path.join(os.path.dirname(__file__), "vcfs", "multi_contig.vcf.gz")
    r = reader.Reader.from_path(path)

    records = [vcf_rec for vcf_rec in r.fetch("20", 1110697, 1234568)]

    assert len(records) == 2

    assert records[0].CHROM == "20"
    assert records[0].POS == 1230237

    assert records[1].CHROM == "20"
    assert records[1].POS == 1234567


# Test fetch with samtools region ---------------------------------------------


def test_fetch_no_records_region():
    path = os.path.join(os.path.dirname(__file__), "vcfs", "multi_contig.vcf.gz")
    r = reader.Reader.from_path(path)

    records = [vcf_rec for vcf_rec in r.fetch("20:1,110,699-1,230,236")]

    assert len(records) == 0


def test_fetch_one_record_region():
    path = os.path.join(os.path.dirname(__file__), "vcfs", "multi_contig.vcf.gz")
    r = reader.Reader.from_path(path)

    records = [vcf_rec for vcf_rec in r.fetch("20:1,110,696-1,230,236")]

    assert len(records) == 1

    assert records[0].CHROM == "20"
    assert records[0].POS == 1110696


def test_fetch_two_records_region():
    path = os.path.join(os.path.dirname(__file__), "vcfs", "multi_contig.vcf.gz")
    r = reader.Reader.from_path(path)

    records = [vcf_rec for vcf_rec in r.fetch("20:1,110,698-1,234,568")]

    assert len(records) == 2

    assert records[0].CHROM == "20"
    assert records[0].POS == 1230237

    assert records[1].CHROM == "20"
    assert records[1].POS == 1234567



# -*- coding: utf-8 -*-
"""Tests for vcfpy.header
"""

import sys

import vcfpy
from vcfpy import header

import pytest


def test_header_field_info():
    """Test the builtin functions of the FieldInfo class"""
    info1 = header.FieldInfo("Integer", 1, "Some description")
    info2 = header.FieldInfo("Integer", 1, "Some description")
    info3 = header.FieldInfo("Integer", ".", "Some description")
    assert info1 == info2
    assert info1 != info3
    assert hash(info1) == hash(info2)
    assert str(info1) == "FieldInfo('Integer', 1, 'Some description', None)"
    assert repr(info1) == "FieldInfo('Integer', 1, 'Some description', None)"


def test_sample_infos():
    info1 = header.SamplesInfos(["one", "two", "three"])
    info2 = header.SamplesInfos(["one", "two", "three"])
    info3 = header.SamplesInfos(["one", "two", "four"])
    assert info1 == info2
    assert info1 != info3
    with pytest.raises(TypeError):
        assert hash(info1)
    assert (
        str(info1)
        == "SamplesInfos(names=['one', 'two', 'three'], name_to_idx={'one': 0, 'three': 2, 'two': 1})"
    )
    assert (
        repr(info1)
        == "SamplesInfos(names=['one', 'two', 'three'], name_to_idx={'one': 0, 'three': 2, 'two': 1})"
    )


def test_header_header():
    lines1 = [header.HeaderLine("foo", "bar"), header.HeaderLine("foo2", "bar2")]
    samples1 = header.SamplesInfos(["one", "two", "three"])
    hdr1 = header.Header(lines1, samples1)

    lines2 = [header.HeaderLine("foo", "bar"), header.HeaderLine("foo2", "bar2")]
    samples2 = header.SamplesInfos(["one", "two", "three"])
    hdr2 = header.Header(lines2, samples2)

    lines3 = [header.HeaderLine("foo3", "bar"), header.HeaderLine("foo2", "bar2")]
    samples3 = header.SamplesInfos(["one", "two", "three"])
    hdr3 = header.Header(lines3, samples3)

    assert hdr1 == hdr2
    assert hdr1 != hdr3
    EXPECTED = (
        "Header(lines=[HeaderLine('foo', 'bar'), HeaderLine('foo2', 'bar2')], "
        "samples=SamplesInfos(names=['one', 'two', 'three'], "
        "name_to_idx={'one': 0, 'three': 2, 'two': 1}))"
    )
    assert str(hdr1) == EXPECTED

    with pytest.raises(TypeError):
        hash(hdr1)


def test_header_without_lines():
    lines = [header.HeaderLine("foo", "bar"), header.HeaderLine("foo2", "bar2")]
    samples = header.SamplesInfos(["one", "two", "three"])
    hdr = header.Header(lines, samples)
    hdr.add_filter_line(vcfpy.OrderedDict([("ID", "PASS")]))
    hdr.add_filter_line(vcfpy.OrderedDict([("ID", "q30")]))
    assert len(hdr.lines) == 4

    hdr2 = header.header_without_lines(hdr, [("foo", "bar"), ("FILTER", "q30")])
    assert len(hdr2.lines) == 2
    assert hdr2.samples == hdr.samples


def test_header_header_line():
    line1 = header.HeaderLine("key", "value")
    line2 = header.HeaderLine("key", "value")
    line3 = header.HeaderLine("key2", "value")
    assert line1 == line2
    assert line1 != line3
    assert str(line1) == "HeaderLine('key', 'value')"
    assert repr(line1) == "HeaderLine('key', 'value')"
    assert line1.value == "value"
    assert line1.serialize() == "##key=value"
    with pytest.raises(TypeError):
        hash(line1)


def test_header_alt_allele_header_line():
    line1 = header.AltAlleleHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "DEL"), ("Description", "deletion")])
    )
    line2 = header.AltAlleleHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "DEL"), ("Description", "deletion")])
    )
    line3 = header.AltAlleleHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "DUP"), ("Description", "duplication")])
    )
    assert line1 == line2
    assert line1 != line3
    if sys.version_info < (3, 6):
        assert str(line1) == (
            """AltAlleleHeaderLine('ALT', '<ID=DEL,Description="deletion">', """
            """OrderedDict([('ID', 'DEL'), ('Description', 'deletion')]))"""
        )
        assert repr(line1) == (
            """AltAlleleHeaderLine('ALT', '<ID=DEL,Description="deletion">', """
            """OrderedDict([('ID', 'DEL'), ('Description', 'deletion')]))"""
        )
    else:
        assert str(line1) == (
            "AltAlleleHeaderLine('ALT', '<ID=DEL,Description=\"deletion\">', "
            "{'ID': 'DEL', 'Description': 'deletion'})"
        )
        assert repr(line1) == (
            "AltAlleleHeaderLine('ALT', '<ID=DEL,Description=\"deletion\">', "
            "{'ID': 'DEL', 'Description': 'deletion'})"
        )
    assert line1.value == '<ID=DEL,Description="deletion">'
    assert line1.serialize() == '##ALT=<ID=DEL,Description="deletion">'
    with pytest.raises(TypeError):
        hash(line1)


def test_header_contig_header_line():
    line1 = header.ContigHeaderLine.from_mapping(vcfpy.OrderedDict([("ID", "1"), ("length", 234)]))
    line2 = header.ContigHeaderLine.from_mapping(vcfpy.OrderedDict([("ID", "1"), ("length", 234)]))
    line3 = header.ContigHeaderLine.from_mapping(vcfpy.OrderedDict([("ID", "2"), ("length", 123)]))
    assert line1 == line2
    assert line1 != line3
    if sys.version_info < (3, 6):
        assert str(line1) == (
            "ContigHeaderLine('contig', '<ID=1,length=234>', OrderedDict([('ID', '1'), ('length', 234)]))"
        )
        assert repr(line1) == (
            "ContigHeaderLine('contig', '<ID=1,length=234>', OrderedDict([('ID', '1'), ('length', 234)]))"
        )
    else:
        assert str(line1) == (
            "ContigHeaderLine('contig', '<ID=1,length=234>', {'ID': '1', 'length': 234})"
        )
        assert repr(line1) == (
            "ContigHeaderLine('contig', '<ID=1,length=234>', {'ID': '1', 'length': 234})"
        )
    assert line1.value == "<ID=1,length=234>"
    assert line1.serialize() == "##contig=<ID=1,length=234>"
    with pytest.raises(TypeError):
        hash(line1)


def test_header_filter_header_line():
    line1 = header.FilterHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "PASS"), ("Description", "All filters passed")])
    )
    line2 = header.FilterHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "PASS"), ("Description", "All filters passed")])
    )
    line3 = header.FilterHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "q30"), ("Description", "Phred score <30")])
    )
    assert line1 == line2
    assert line1 != line3
    if sys.version_info < (3, 6):
        assert str(line1) == (
            "FilterHeaderLine('FILTER', '<ID=PASS,Description=\"All filters passed\">', "
            "OrderedDict([('ID', 'PASS'), ('Description', 'All filters passed')]))"
        )
        assert repr(line1) == (
            "FilterHeaderLine('FILTER', '<ID=PASS,Description=\"All filters passed\">', "
            "OrderedDict([('ID', 'PASS'), ('Description', 'All filters passed')]))"
        )
    else:
        assert str(line1) == (
            "FilterHeaderLine('FILTER', '<ID=PASS,Description=\"All filters passed\">', "
            "{'ID': 'PASS', 'Description': 'All filters passed'})"
        )
        assert repr(line1) == (
            "FilterHeaderLine('FILTER', '<ID=PASS,Description=\"All filters passed\">', "
            "{'ID': 'PASS', 'Description': 'All filters passed'})"
        )
    assert line1.value == '<ID=PASS,Description="All filters passed">'
    assert line1.serialize() == '##FILTER=<ID=PASS,Description="All filters passed">'
    with pytest.raises(TypeError):
        hash(line1)


def test_header_pedigree_header_line():
    line1 = header.PedigreeHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "child"), ("Father", "father")])
    )
    line2 = header.PedigreeHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "child"), ("Father", "father")])
    )
    line3 = header.PedigreeHeaderLine.from_mapping(vcfpy.OrderedDict([("ID", "father")]))
    assert line1 == line2
    assert line1 != line3
    if sys.version_info < (3, 6):
        assert str(line1) == (
            "PedigreeHeaderLine('PEDIGREE', '<ID=child,Father=father>', "
            "OrderedDict([('ID', 'child'), ('Father', 'father')]))"
        )
        assert repr(line1) == (
            "PedigreeHeaderLine('PEDIGREE', '<ID=child,Father=father>', "
            "OrderedDict([('ID', 'child'), ('Father', 'father')]))"
        )
    else:
        assert str(line1) == (
            "PedigreeHeaderLine('PEDIGREE', '<ID=child,Father=father>', {'ID': 'child', 'Father': 'father'})"
        )
        assert repr(line1) == (
            "PedigreeHeaderLine('PEDIGREE', '<ID=child,Father=father>', {'ID': 'child', 'Father': 'father'})"
        )
    assert line1.value == "<ID=child,Father=father>"
    assert line1.serialize() == "##PEDIGREE=<ID=child,Father=father>"
    with pytest.raises(TypeError):
        hash(line1)


def test_header_sample_header_line():
    line1 = header.SampleHeaderLine.from_mapping(vcfpy.OrderedDict([("ID", "sample1")]))
    line2 = header.SampleHeaderLine.from_mapping(vcfpy.OrderedDict([("ID", "sample1")]))
    line3 = header.SampleHeaderLine.from_mapping(vcfpy.OrderedDict([("ID", "sample2")]))
    assert line1 == line2
    assert line1 != line3
    if sys.version_info < (3, 6):
        assert str(line1) == (
            "SampleHeaderLine('SAMPLE', '<ID=sample1>', OrderedDict([('ID', 'sample1')]))"
        )
        assert repr(line1) == (
            "SampleHeaderLine('SAMPLE', '<ID=sample1>', OrderedDict([('ID', 'sample1')]))"
        )
    else:
        assert str(line1) == ("SampleHeaderLine('SAMPLE', '<ID=sample1>', {'ID': 'sample1'})")
        assert repr(line1) == ("SampleHeaderLine('SAMPLE', '<ID=sample1>', {'ID': 'sample1'})")
    assert line1.value == "<ID=sample1>"
    assert line1.serialize() == "##SAMPLE=<ID=sample1>"
    with pytest.raises(TypeError):
        hash(line1)


def test_header_info_header_line():
    line1 = header.InfoHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "SVTYPE"), ("Number", 1), ("Type", "String")])
    )
    line2 = header.InfoHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "SVTYPE"), ("Number", 1), ("Type", "String")])
    )
    line3 = header.InfoHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "END"), ("Number", 1), ("Type", "Integer")])
    )
    assert line1 == line2
    assert line1 != line3
    if sys.version_info < (3, 6):
        assert str(line1) == (
            "InfoHeaderLine('INFO', '<ID=SVTYPE,Number=1,Type=String>', "
            "OrderedDict([('ID', 'SVTYPE'), ('Number', 1), ('Type', 'String')]))"
        )
        assert repr(line1) == (
            "InfoHeaderLine('INFO', '<ID=SVTYPE,Number=1,Type=String>', "
            "OrderedDict([('ID', 'SVTYPE'), ('Number', 1), ('Type', 'String')]))"
        )
    else:
        assert str(line1) == (
            "InfoHeaderLine('INFO', '<ID=SVTYPE,Number=1,Type=String>', "
            "{'ID': 'SVTYPE', 'Number': 1, 'Type': 'String'})"
        )
        assert repr(line1) == (
            "InfoHeaderLine('INFO', '<ID=SVTYPE,Number=1,Type=String>', "
            "{'ID': 'SVTYPE', 'Number': 1, 'Type': 'String'})"
        )
    assert line1.value == "<ID=SVTYPE,Number=1,Type=String>"
    assert line1.serialize() == "##INFO=<ID=SVTYPE,Number=1,Type=String>"
    with pytest.raises(TypeError):
        hash(line1)


def test_header_format_header_line():
    line1 = header.FormatHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "AD"), ("Number", "R"), ("Type", "Integer")])
    )
    line2 = header.FormatHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "AD"), ("Number", "R"), ("Type", "Integer")])
    )
    line3 = header.FormatHeaderLine.from_mapping(
        vcfpy.OrderedDict([("ID", "DP"), ("Number", 1), ("Type", "Integer")])
    )
    assert line1 == line2
    assert line1 != line3
    if sys.version_info < (3, 6):
        assert str(line1) == (
            "FormatHeaderLine('FORMAT', '<ID=AD,Number=R,Type=Integer>', "
            "OrderedDict([('ID', 'AD'), ('Number', 'R'), ('Type', 'Integer')]))"
        )
        assert repr(line1) == (
            "FormatHeaderLine('FORMAT', '<ID=AD,Number=R,Type=Integer>', "
            "OrderedDict([('ID', 'AD'), ('Number', 'R'), ('Type', 'Integer')]))"
        )
    else:
        assert str(line1) == (
            "FormatHeaderLine('FORMAT', '<ID=AD,Number=R,Type=Integer>', "
            "{'ID': 'AD', 'Number': 'R', 'Type': 'Integer'})"
        )
        assert repr(line1) == (
            "FormatHeaderLine('FORMAT', '<ID=AD,Number=R,Type=Integer>', "
            "{'ID': 'AD', 'Number': 'R', 'Type': 'Integer'})"
        )
    assert line1.value == "<ID=AD,Number=R,Type=Integer>"
    assert line1.serialize() == "##FORMAT=<ID=AD,Number=R,Type=Integer>"
    with pytest.raises(TypeError):
        hash(line1)


def test_header_has_header_line_positive():
    lines = [
        header.FormatHeaderLine.from_mapping(
            vcfpy.OrderedDict([("ID", "DP"), ("Number", "R"), ("Type", "Integer")])
        ),
        header.InfoHeaderLine.from_mapping(
            vcfpy.OrderedDict([("ID", "AD"), ("Number", "R"), ("Type", "Integer")])
        ),
        header.FilterHeaderLine.from_mapping(
            vcfpy.OrderedDict([("ID", "PASS"), ("Description", "All filters passed")])
        ),
        header.ContigHeaderLine.from_mapping(vcfpy.OrderedDict([("ID", "1"), ("length", 234)])),
    ]
    samples = header.SamplesInfos(["one", "two", "three"])
    hdr = header.Header(lines, samples)

    assert hdr.has_header_line("FORMAT", "DP")
    assert hdr.has_header_line("INFO", "AD")
    assert hdr.has_header_line("FILTER", "PASS")
    assert hdr.has_header_line("contig", "1")


def test_header_has_header_line_positive_no_samples():
    lines = []
    samples = header.SamplesInfos(["one", "two", "three"])
    hdr = header.Header(lines, samples)

    assert not hdr.has_header_line("FORMAT", "DP")
    assert not hdr.has_header_line("INFO", "AD")
    assert not hdr.has_header_line("FILTER", "PASS")
    assert not hdr.has_header_line("contig", "1")



# -*- coding: utf-8 -*-
"""Writing of VCF records
"""

import io
import textwrap

import pytest

import vcfpy
from vcfpy import parser, writer, header, record

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"

# TODO: cleanup, refactor tests somewhat

MEDIUM_HEADER = """
##fileformat=VCFv4.3
##fileDate=20090805
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##INFO=<ID=ANNO,Number=.,Type=String,Description="Additional annotation">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
##FORMAT=<ID=FT,Number=.,Type=String,Description="Call-wise filters">
#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00001\tNA00002\tNA00003
""".lstrip()


@pytest.fixture(scope="function")
def header_samples():
    p = parser.Parser(stream=io.StringIO(MEDIUM_HEADER), path="<builtin>")
    p.parse_header()
    return (p.header, p.samples)


def test_write_minimal_record(header_samples, tmpdir_factory):
    O = vcfpy.OrderedDict
    # open temporary file and setup the Writer with header
    path = tmpdir_factory.mktemp("write_header").join("out.vcf")
    header, _ = header_samples
    w = writer.Writer.from_path(path, header)
    # construct record to write out from scratch
    r = record.Record(
        "20",
        100,
        [],
        "C",
        [record.Substitution(record.SNV, "T")],
        None,
        [],
        O(),
        ["GT"],
        [
            record.Call("NA00001", O(GT="0/1")),
            record.Call("NA00002", O(GT="0/0")),
            record.Call("NA00003", O(GT="1/1")),
        ],
    )
    # write out the record, close file to ensure flushing to disk
    w.write_record(r)
    w.close()
    # compare actual result with expected
    RESULT = path.read()
    LINE = "20\t100\t.\tC\tT\t.\t.\t.\tGT\t0/1\t0/0\t1/1\n"
    EXPECTED = MEDIUM_HEADER + LINE
    assert EXPECTED == RESULT


def test_write_annotated_record(header_samples, tmpdir_factory):
    O = vcfpy.OrderedDict
    S = record.Substitution
    # open temporary file and setup the Writer with header
    path = tmpdir_factory.mktemp("write_annotated_record").join("out.vcf")
    header, _ = header_samples
    w = writer.Writer.from_path(path, header)
    # construct record to write out from scratch
    r = record.Record(
        "20",
        100,
        ["rs333", "CSN42"],
        "C",
        [record.Substitution(record.SNV, "T"), record.Substitution(record.SNV, "G")],
        50,
        ["PASS"],
        O([("DP", 93), ("AF", [0.3, 0.2]), ("DB", True)]),
        ["GT", "DP", "GQ", "HQ"],
        [
            record.Call("NA00001", O(GT="0/1", DP=30, GQ=40, HQ=[1, 2])),
            record.Call("NA00002", O(GT="0/2", DP=31, GQ=41, HQ=[3, 4])),
            record.Call("NA00003", O(GT="1/2", DP=32, GQ=42, HQ=[5, 6])),
        ],
    )
    # write out the record, close file to ensure flushing to disk
    w.write_record(r)
    w.close()
    # compare actual result with expected
    RESULT = path.read()
    LINE = "20\t100\trs333;CSN42\tC\tT,G\t50\tPASS\tDP=93;AF=0.3,0.2;DB\tGT:DP:GQ:HQ\t0/1:30:40:1,2\t0/2:31:41:3,4\t1/2:32:42:5,6\n"
    EXPECTED = MEDIUM_HEADER + LINE
    assert EXPECTED == RESULT


def test_write_record_with_escaping(header_samples, tmpdir_factory):
    O = vcfpy.OrderedDict
    S = record.Substitution
    # open temporary file and setup the Writer with header
    path = tmpdir_factory.mktemp("write_header").join("out.vcf")
    header, _ = header_samples
    w = writer.Writer.from_path(path, header)
    # construct record to write out from scratch
    r = record.Record(
        "20",
        100,
        [],
        "C",
        [record.Substitution(record.SNV, "T")],
        None,
        [],
        O([("ANNO", ["Here,are%some chars", "%25"])]),
        ["GT", "FT"],
        [
            record.Call("NA00001", O(GT="0/1", FT=["%25", "FOO"])),
            record.Call("NA00002", O(GT="0/0", FT=[])),
            record.Call("NA00003", O(GT="1/1", FT=[])),
        ],
    )
    # write out the record, close file to ensure flushing to disk
    w.write_record(r)
    w.close()
    # compare actual result with expected
    RESULT = path.read()
    LINE = (
        "20\t100\t.\tC\tT\t.\t.\tANNO=Here%2Care%25some chars,"
        "%2525\tGT:FT\t0/1:%2525;FOO\t0/0:.\t1/1:.\n"
    )
    EXPECTED = MEDIUM_HEADER + LINE
    assert EXPECTED == RESULT


def test_write_record_no_samples(tmpdir_factory):
    O = vcfpy.OrderedDict
    # Create header without samples
    hdr = header.Header(
        lines=[header.HeaderLine("fileformat", "VCFv4.0")], samples=header.SamplesInfos([])
    )
    # construct record to write out from scratch
    r = record.Record("20", 100, [], "C", [record.Substitution(record.SNV, "T")], None, [], O())
    # Write out header and record
    path = tmpdir_factory.mktemp("write_header").join("out.vcf")
    w = writer.Writer.from_path(path, hdr)
    w.write_record(r)
    w.close()
    # Compare result
    RESULT = path.read()
    EXPECTED = textwrap.dedent(
        """
    ##fileformat=VCFv4.0
    #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO
    20\t100\t.\tC\tT\t.\t.\t.
    """
    ).lstrip()
    assert RESULT == EXPECTED



# -*- coding: utf-8 -*-
"""Exceptions for the vcfpy module
"""

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


class VCFPyException(RuntimeError):
    """Base class for module's exception"""


class InvalidHeaderException(VCFPyException):
    """Raised in the case of invalid header formatting"""


class InvalidRecordException(VCFPyException):
    """Raised in the case of invalid record formatting"""


class IncorrectVCFFormat(VCFPyException):
    """Raised on problems parsing VCF"""


class HeaderNotFound(VCFPyException):
    """Raised when a VCF header could not be found"""


class VCFPyWarning(Warning):
    """Base class for module's warnings"""


class DuplicateHeaderLineWarning(VCFPyWarning):
    """A header line occurs twice in a header"""


class FieldInfoNotFound(VCFPyWarning):
    """A header field is not found, default is used"""


class FieldMissingNumber(VCFPyWarning):
    """Raised when compound heade misses number"""


class FieldInvalidNumber(VCFPyWarning):
    """Raised when compound header has invalid number"""


class HeaderInvalidType(VCFPyWarning):
    """Raised when compound header has invalid type"""


class HeaderMissingDescription(VCFPyWarning):
    """Raised when compound header has missing description"""


class LeadingTrailingSpaceInKey(VCFPyWarning):
    """Leading or trailing space in key"""


class UnknownFilter(VCFPyWarning):
    """Missing FILTER"""


class UnknownVCFVersion(VCFPyWarning):
    """Unknown VCF version"""


class IncorrectListLength(VCFPyWarning):
    """Wrong length of multi-element field"""


class SpaceInChromLine(VCFPyWarning):
    """Space instead of TAB in ##CHROM line"""


class CannotConvertValue(VCFPyWarning):
    """Cannot convert value."""



# -*- coding: utf-8 -*-
"""Support code for writing BGZF files

Shamelessly taken from Biopython
"""

#                 Biopython License Agreement
#
# Permission to use, copy, modify, and distribute this software and its
# documentation with or without modifications and for any purpose and
# without fee is hereby granted, provided that any copyright notices
# appear in all copies and that both those copyright notices and this
# permission notice appear in supporting documentation, and that the
# names of the contributors or copyright holders not be used in
# advertising or publicity pertaining to distribution of the software
# without specific prior permission.
#
# THE CONTRIBUTORS AND COPYRIGHT HOLDERS OF THIS SOFTWARE DISCLAIM ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
# OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
# OR PERFORMANCE OF THIS SOFTWARE.

import codecs
import struct
import zlib


# For Python 2 can just use: _bgzf_magic = '\x1f\x8b\x08\x04'
# but need to use bytes on Python 3
_bgzf_magic = b"\x1f\x8b\x08\x04"
_bgzf_header = b"\x1f\x8b\x08\x04\x00\x00\x00\x00\x00\xff\x06\x00\x42" b"\x43\x02\x00"
_bgzf_eof = (
    b"\x1f\x8b\x08\x04\x00\x00\x00\x00\x00\xff\x06\x00BC\x02\x00"
    b"\x1b\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00"
)
_bytes_BC = b"BC"


def make_virtual_offset(block_start_offset, within_block_offset):
    """Compute a BGZF virtual offset from block start and within block offsets.
    The BAM indexing scheme records read positions using a 64 bit
    'virtual offset', comprising in C terms:
    block_start_offset << 16 | within_block_offset
    Here block_start_offset is the file offset of the BGZF block
    start (unsigned integer using up to 64-16 = 48 bits), and
    within_block_offset within the (decompressed) block (unsigned
    16 bit integer).
    >>> make_virtual_offset(0, 0)
    0
    >>> make_virtual_offset(0, 1)
    1
    >>> make_virtual_offset(0, 2**16 - 1)
    65535
    >>> make_virtual_offset(0, 2**16)
    Traceback (most recent call last):
    ...
    ValueError: Require 0 <= within_block_offset < 2**16, got 65536
    >>> 65536 == make_virtual_offset(1, 0)
    True
    >>> 65537 == make_virtual_offset(1, 1)
    True
    >>> 131071 == make_virtual_offset(1, 2**16 - 1)
    True
    >>> 6553600000 == make_virtual_offset(100000, 0)
    True
    >>> 6553600001 == make_virtual_offset(100000, 1)
    True
    >>> 6553600010 == make_virtual_offset(100000, 10)
    True
    >>> make_virtual_offset(2**48, 0)
    Traceback (most recent call last):
    ...
    ValueError: Require 0 <= block_start_offset < 2**48, got 281474976710656
    """
    if within_block_offset < 0 or within_block_offset >= 65536:
        raise ValueError("Require 0 <= within_block_offset < 2**16, got %i" % within_block_offset)
    if block_start_offset < 0 or block_start_offset >= 281474976710656:
        raise ValueError("Require 0 <= block_start_offset < 2**48, got %i" % block_start_offset)
    return (block_start_offset << 16) | within_block_offset


class BgzfWriter(object):
    def __init__(self, filename=None, mode="w", fileobj=None, compresslevel=6):
        if fileobj:
            assert filename is None
            handle = fileobj
        else:
            if "w" not in mode.lower() and "a" not in mode.lower():
                raise ValueError("Must use write or append mode, not %r" % mode)
            if "a" in mode.lower():
                handle = open(filename, "ab")
            else:
                handle = open(filename, "wb")
        self._text = "b" not in mode.lower()
        self._handle = handle
        self._buffer = b""
        self.compresslevel = compresslevel

    def _write_block(self, block):
        # print("Saving %i bytes" % len(block))
        assert len(block) <= 65536
        # Giving a negative window bits means no gzip/zlib headers,
        # -15 used in samtools
        c = zlib.compressobj(self.compresslevel, zlib.DEFLATED, -15, zlib.DEF_MEM_LEVEL, 0)
        compressed = c.compress(block) + c.flush()
        del c
        assert len(compressed) < 65536, "TODO - Didn't compress enough, try less data in this block"
        crc = zlib.crc32(block)
        # Should cope with a mix of Python platforms...
        if crc < 0:
            crc = struct.pack("<i", crc)
        else:
            crc = struct.pack("<I", crc)
        bsize = struct.pack("<H", len(compressed) + 25)  # includes -1
        crc = struct.pack("<I", zlib.crc32(block) & 0xFFFFFFFF)
        uncompressed_length = struct.pack("<I", len(block))
        # Fixed 16 bytes,
        # gzip magic bytes (4) mod time (4),
        # gzip flag (1), os (1), extra length which is six (2),
        # sub field which is BC (2), sub field length of two (2),
        # Variable data,
        # 2 bytes: block length as BC sub field (2)
        # X bytes: the data
        # 8 bytes: crc (4), uncompressed data length (4)
        data = _bgzf_header + bsize + compressed + crc + uncompressed_length
        self._handle.write(data)

    def write(self, data):
        # TODO - Check bytes vs unicode
        if isinstance(data, str):
            data = codecs.latin_1_encode(data)[0]
        # block_size = 2**16 = 65536
        data_len = len(data)
        if len(self._buffer) + data_len < 65536:
            # print("Cached %r" % data)
            self._buffer += data
            return
        else:
            # print("Got %r, writing out some data..." % data)
            self._buffer += data
            while len(self._buffer) >= 65536:
                self._write_block(self._buffer[:65536])
                self._buffer = self._buffer[65536:]

    def flush(self):
        while len(self._buffer) >= 65536:
            self._write_block(self._buffer[:65535])
            self._buffer = self._buffer[65535:]
        self._write_block(self._buffer)
        self._buffer = b""
        self._handle.flush()

    def close(self):
        """Flush data, write 28 bytes BGZF EOF marker, and close BGZF file.
        samtools will look for a magic EOF marker, just a 28 byte empty BGZF
        block, and if it is missing warns the BAM file may be truncated. In
        addition to samtools writing this block, so too does bgzip - so this
        implementation does too.
        """
        if self._buffer:
            self.flush()
        self._handle.write(_bgzf_eof)
        self._handle.flush()
        self._handle.close()

    def tell(self):
        """Returns a BGZF 64-bit virtual offset."""
        return make_virtual_offset(self._handle.tell(), len(self._buffer))

    def seekable(self):
        # Not seekable, but we do support tell...
        return False

    @classmethod
    def isatty(klass):
        return False

    def fileno(self):
        return self._handle.fileno()

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        self.close()



# -*- coding: utf-8 -*-
"""Writing of VCF files to ``file``-like objects

Currently, only writing to plain-text files is supported
"""

from . import parser
from . import record
from . import bgzf

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


def format_atomic(value, section):
    """Format atomic value

    This function also takes care of escaping the value in case one of the
    reserved characters occurs in the value.
    """
    # Perform escaping
    if isinstance(value, str):
        if any(r in value for r in record.RESERVED_CHARS[section]):
            for k, v in record.ESCAPE_MAPPING:
                value = value.replace(k, v)
    # String-format the given value
    if value is None:
        return "."
    else:
        return str(value)


def format_value(field_info, value, section):
    """Format possibly compound value given the FieldInfo"""
    if section == "FORMAT" and field_info.id == "FT":
        if not value:
            return "."
        elif isinstance(value, list):
            return ";".join(map(lambda x: format_atomic(x, section), value))
    elif field_info.number == 1:
        if value is None:
            return "."
        else:
            return format_atomic(value, section)
    else:
        if not value:
            return "."
        else:
            return ",".join(map(lambda x: format_atomic(x, section), value))


class Writer:
    """Class for writing VCF files to ``file``-like objects

    Instead of using the constructor, use the class methods
    :py:meth:`~Writer.from_stream` and
    :py:meth:`~Writer.from_path`.

    The writer has to be constructed with a :py:class:`~vcfpy.header.Header`
    object and the full VCF header will be written immediately on construction.
    This, of course, implies that modifying the header after construction is
    illegal.
    """

    @classmethod
    def from_stream(klass, stream, header, path=None, use_bgzf=None):
        """Create new :py:class:`Writer` from file

        Note that for getting bgzf support, you have to pass in a stream
        opened in binary mode.  Further, you either have to provide a ``path``
        ending in ``".gz"`` or set ``use_bgzf=True``.  Otherwise, you will
        get the notorious "TypeError: 'str' does not support the buffer
        interface".

        :param stream: ``file``-like object to write to
        :param header: VCF header to use, lines and samples are deep-copied
        :param path: optional string with path to store (for display only)
        :param use_bgzf: indicator whether to write bgzf to ``stream``
            if ``True``, prevent if ``False``, interpret ``path`` if ``None``
        """
        if use_bgzf or (use_bgzf is None and path and path.endswith(".gz")):
            stream = bgzf.BgzfWriter(fileobj=stream)
        return Writer(stream, header, path)

    @classmethod
    def from_path(klass, path, header):
        """Create new :py:class:`Writer` from path

        :param path: the path to load from (converted to ``str`` for
            compatibility with ``path.py``)
        :param header: VCF header to use, lines and samples are deep-copied
        """
        path = str(path)
        use_bgzf = False  # we already interpret path
        if path.endswith(".gz"):
            f = bgzf.BgzfWriter(filename=path)
        else:
            f = open(path, "wt")
        return klass.from_stream(f, header, path, use_bgzf=use_bgzf)

    def __init__(self, stream, header, path=None):
        #: stream (``file``-like object) to read from
        self.stream = stream
        #: the :py:class:~vcfpy.header.Header` to write out, will be
        #: deep-copied into the ``Writer`` on initialization
        self.header = header.copy()
        #: optional ``str`` with the path to the stream
        self.path = path
        # write out headers
        self._write_header()

    def _write_header(self):
        """Write out the header"""
        for line in self.header.lines:
            print(line.serialize(), file=self.stream)
        if self.header.samples.names:
            print(
                "\t".join(list(parser.REQUIRE_SAMPLE_HEADER) + self.header.samples.names),
                file=self.stream,
            )
        else:
            print("\t".join(parser.REQUIRE_NO_SAMPLE_HEADER), file=self.stream)

    def close(self):
        """Close underlying stream"""
        self.stream.close()

    def write_record(self, record):
        """Write out the given :py:class:`vcfpy.record.Record` to this
        Writer"""
        self._serialize_record(record)

    def _serialize_record(self, record):
        """Serialize whole Record"""
        f = self._empty_to_dot
        row = [record.CHROM, record.POS]
        row.append(f(";".join(record.ID)))
        row.append(f(record.REF))
        if not record.ALT:
            row.append(".")
        else:
            row.append(",".join([f(a.serialize()) for a in record.ALT]))
        row.append(f(record.QUAL))
        row.append(f(";".join(record.FILTER)))
        row.append(f(self._serialize_info(record)))
        if record.FORMAT:
            row.append(":".join(record.FORMAT))
        row += [
            self._serialize_call(record.FORMAT, record.call_for_sample[s])
            for s in self.header.samples.names
        ]
        print(*row, sep="\t", file=self.stream)

    def _serialize_info(self, record):
        """Return serialized version of record.INFO"""
        result = []
        for key, value in record.INFO.items():
            info = self.header.get_info_field_info(key)
            if info.type == "Flag":
                result.append(key)
            else:
                result.append("{}={}".format(key, format_value(info, value, "INFO")))
        return ";".join(result)

    def _serialize_call(self, format_, call):
        """Return serialized version of the Call using the record's FORMAT'"""
        if isinstance(call, record.UnparsedCall):
            return call.unparsed_data
        else:
            result = [
                format_value(self.header.get_format_field_info(key), call.data.get(key), "FORMAT")
                for key in format_
            ]
            return ":".join(result)

    @classmethod
    def _empty_to_dot(klass, val):
        """Return val or '.' if empty value"""
        if val == "" or val is None or val == []:
            return "."
        else:
            return val

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        self.close()



# -*- coding: utf-8 -*-
"""Tests for reading writing samples in different order from reading them
"""

from vcfpy import Reader, Writer, Header, SamplesInfos


def test_reading_and_write_reordered(tmpdir, multisample_vcf_file, multisample_vcf_reordered):
    # Perform record-wise copying, saving results in records, not writing
    # out for some samples
    records = []
    out_path = str(tmpdir.mkdir("output").join("output.vcf"))
    with Reader.from_path(multisample_vcf_file) as reader:
        samples = SamplesInfos(["NA00002", "NA00003", "NA00001"])
        header = Header(reader.header.lines, samples)
        with Writer.from_path(out_path, header) as writer:
            for record in reader:
                records.append(record)
                writer.write_record(record)
    # Check resulting file
    with open(out_path, "rt") as outf:
        assert multisample_vcf_reordered == outf.read()


def test_reading_and_write_reordered_parse_subset(
    tmpdir, multisample_vcf_file, multisample_vcf_reordered
):
    # Perform record-wise copying, saving results in records, not writing
    # out for some samples
    records = []
    out_path = str(tmpdir.mkdir("output").join("output.vcf"))
    with Reader.from_path(multisample_vcf_file, parsed_samples=["NA00001"]) as reader:
        samples = SamplesInfos(["NA00002", "NA00003", "NA00001"])
        header = Header(reader.header.lines, samples)
        with Writer.from_path(out_path, header) as writer:
            for record in reader:
                records.append(record)
                writer.write_record(record)
    # Check resulting file
    with open(out_path, "rt") as outf:
        assert multisample_vcf_reordered == outf.read()



# -*- coding: utf-8 -*-
"""Code for defining types depending on fast implementations

Namely, defines the type ``OrderedDict``:

- on CPython >=3.6, define as ``dict``
- otherwise, if ``cyordereddict`` is available use
  ``cyordereddict.OrderedDict``
- otherwise, use ``collections.OrderedDict``
"""

import sys


if sys.version_info[:2] >= (3, 6):
    OrderedDict = dict
else:
    try:
        from cyordereddict import OrderedDict
    except ImportError:
        from collections import OrderedDict  # noqa: ignore=F401



# -*- coding: utf-8 -*-
"""Tests for the Reader and Writer working as context managers
"""

import io
import os

import pytest

import vcfpy
from vcfpy import reader
from vcfpy import writer
from vcfpy import parser
from vcfpy import record

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


MEDIUM_HEADER = """
##fileformat=VCFv4.3
##fileDate=20090805
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
##ALT=<ID=R,Description="IUPAC code R = A/G">
##META=<ID=Assay,Type=String,Number=.,Values=[WholeGenome, Exome]>
##PEDIGREE=<ID=TumourSample,Original=GermlineID>
##SAMPLE=<ID=Sample1,Assay=WholeGenome,Ethnicity=AFR,Disease=None,Description="Patient germline genome from unaffected",DOI=url>
#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00001\tNA00002\tNA00003
""".lstrip()


@pytest.fixture(scope="function")
def header_samples():
    p = parser.Parser(stream=io.StringIO(MEDIUM_HEADER), path="<builtin>")
    p.parse_header()
    return (p.header, p.samples)


def test_reader():
    path = os.path.join(os.path.dirname(__file__), "vcfs/full_vcf43.vcf")
    with reader.Reader.from_path(path) as r:
        # should be open now
        assert r.stream
        assert not r.stream.closed
        # read records
        records = [rec for rec in r]
    # should be closed now
    assert r.stream
    assert r.stream.closed
    # check result
    assert len(records) == 5


def test_reader_fetch():
    path = os.path.join(os.path.dirname(__file__), "vcfs", "multi_contig.vcf.gz")
    with reader.Reader.from_path(path) as r:
        records = [vcf_rec for vcf_rec in r.fetch("20", 1110695, 1230236)]
        assert r.stream
        assert not r.stream.closed
        assert r.tabix_file
        assert not r.tabix_file.closed
    # closed
    assert r.stream
    assert r.stream.closed
    assert r.tabix_file
    assert r.tabix_file.closed
    # check result
    assert len(records) == 1
    assert records[0].CHROM == "20"
    assert records[0].POS == 1110696


def test_writer(header_samples, tmpdir_factory):
    O = vcfpy.OrderedDict
    # open temporary file and setup the Writer with header
    path = tmpdir_factory.mktemp("write_header").join("out.vcf")
    header, _ = header_samples
    # construct record to write out from scratch
    r = record.Record(
        "20",
        100,
        [],
        "C",
        [record.Substitution(record.SNV, "T")],
        None,
        [],
        O(),
        ["GT"],
        [
            record.Call("NA00001", O(GT="0/1")),
            record.Call("NA00002", O(GT="0/0")),
            record.Call("NA00003", O(GT="1/1")),
        ],
    )
    # open writer
    with writer.Writer.from_path(path, header) as w:
        # write out the record
        w.write_record(r)
    # should be closed
    assert w.stream
    assert w.stream.closed
    # compare actual result with expected
    RESULT = path.read()
    LINE = "20\t100\t.\tC\tT\t.\t.\t.\tGT\t0/1\t0/0\t1/1\n"
    EXPECTED = MEDIUM_HEADER + LINE
    assert EXPECTED == RESULT



# -*- coding: utf-8 -*-
"""Test the Record class basics."""

import sys

import vcfpy


def test_record_from_scratch():
    """Test contruction of Record objectss"""
    record = vcfpy.Record(
        CHROM="chr1",
        POS=1234,
        ID=[],
        REF="A",
        ALT=[vcfpy.Substitution("SNV", "T")],
        QUAL=None,
        FILTER=[],
        INFO=vcfpy.OrderedDict(),
    )
    record.calls = [vcfpy.Call("sample-1", vcfpy.OrderedDict())]
    record.add_format("GT", "./.")
    if sys.version_info < (3, 6):
        assert str(record) == (
            "Record('chr1', 1234, [], 'A', [Substitution(type_='SNV', value='T')], None, "
            "[], OrderedDict(), ['GT'], [Call('sample-1', OrderedDict([('GT', './.')]))])"
        )
    else:
        assert str(record) == (
            "Record('chr1', 1234, [], 'A', [Substitution(type_='SNV', value='T')], None, "
            "[], {}, ['GT'], [Call('sample-1', {'GT': './.'})])"
        )



# -*- coding: utf-8 -*-
"""Test Call class
"""

import vcfpy
from vcfpy import record

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


def build_rec(calls=None, format_extras=None):
    calls = calls or []
    format_extras = format_extras or []
    alt1 = record.Substitution(vcfpy.SNV, "T")
    alt2 = record.Substitution(vcfpy.SNV, "A")
    return record.Record(
        "2",
        100,
        [],
        "C",
        [alt1, alt2],
        None,
        [],
        vcfpy.OrderedDict(),
        ["GT"] + format_extras,
        calls,
    )


# Call.is_phased() ------------------------------------------------------------


def test_is_phased_true():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0|1")]))
    assert call.is_phased is True


def test_is_phased_mixed():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0/1|2")]))
    assert call.is_phased is True


def test_is_phased_false():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0/1")]))
    assert call.is_phased is False


# Call.gt_phase_char() --------------------------------------------------------


def test_gt_phase_char_pipe():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0|1")]))
    assert call.gt_phase_char == "|"


def test_gt_phase_char_slash():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0/1")]))
    assert call.gt_phase_char == "/"


# Call.gt_bases() -------------------------------------------------------------


def test_gt_bases_0_0():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0|0")]))
    build_rec([call])
    assert call.gt_bases == ("C", "C")


def test_gt_bases_0_1():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0|1")]))
    build_rec([call])
    assert call.gt_bases == ("C", "T")


def test_gt_bases_1_1():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "1|1")]))
    build_rec([call])
    assert call.gt_bases == ("T", "T")


def test_gt_bases_0_2():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0|2")]))
    build_rec([call])
    assert call.gt_bases == ("C", "A")


# Call.gt_type() --------------------------------------------------------------


def test_gt_type_het():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0|1")]))
    assert call.gt_type == vcfpy.HET


def test_gt_type_hom_ref():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0/0")]))
    assert call.gt_type == vcfpy.HOM_REF


def test_gt_type_hom_alt():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "1/1")]))
    assert call.gt_type == vcfpy.HOM_ALT


# Call.is_het() ---------------------------------------------------------------


def test_is_het_het():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0|1")]))
    assert call.is_het


def test_is_het_hom_ref():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0/0")]))
    assert not call.is_het


def test_is_het_hom_alt():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "1/1")]))
    assert not call.is_het


# Call.is_variant() -----------------------------------------------------------


def test_is_variant_het():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0|1")]))
    assert call.is_variant


def test_is_variant_hom_ref():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "0/0")]))
    assert not call.is_variant


def test_is_variant_hom_alt():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "1/1")]))
    assert call.is_variant


def test_is_variant_no_call():
    call1 = record.Call("sample", vcfpy.OrderedDict([("GT", ".")]))
    assert not call1.is_variant
    call2 = record.Call("sample", vcfpy.OrderedDict([("GT", "./.")]))
    assert not call2.is_variant


# Call.ploidy ----------------------------------------------------------------


def test_ploidy_nocall():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", ".")]))
    assert call.ploidy == 1


def test_ploidy_one():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "1")]))
    assert call.ploidy == 1


def test_ploidy_two():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "1/1")]))
    assert call.ploidy == 2


# Call.is_filtered() ----------------------------------------------------------


def test_gt_type_filtered_no_ft():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "1/1")]))
    assert not call.is_filtered()


def test_gt_type_filtered_empty():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "1/1"), ("FT", [])]))
    assert not call.is_filtered()


def test_gt_type_filtered_pass():
    call = record.Call("sample", vcfpy.OrderedDict([("GT", "1/1"), ("FT", ["PASS"])]))
    assert not call.is_filtered()



# -*- coding: utf-8 -*-
"""Code for representing the VCF header part

The VCF header class structure is modeled after HTSJDK
"""

import json
import pprint
import warnings

from . import exceptions
from .compat import OrderedDict
from .exceptions import (
    DuplicateHeaderLineWarning,
    FieldInfoNotFound,
    FieldMissingNumber,
    FieldInvalidNumber,
    HeaderInvalidType,
    HeaderMissingDescription,
)

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"

# Tuples of valid entries -----------------------------------------------------
#
#: valid INFO value types
INFO_TYPES = ("Integer", "Float", "Flag", "Character", "String")
#: valid FORMAT value types
FORMAT_TYPES = ("Integer", "Float", "Character", "String")
#: valid values for "Number" entries, except for integers
VALID_NUMBERS = ("A", "R", "G", ".")
#: header lines that contain an "ID" entry
LINES_WITH_ID = ("ALT", "contig", "FILTER", "FORMAT", "INFO", "META", "PEDIGREE", "SAMPLE")

# Constants for "Number" entries ----------------------------------------------
#
#: number of alleles excluding reference
HEADER_NUMBER_ALLELES = "A"
#: number of alleles including reference
HEADER_NUMBER_REF = "R"
#: number of genotypes
HEADER_NUMBER_GENOTYPES = "G"
#: unbounded number of values
HEADER_NUMBER_UNBOUNDED = "."


class FieldInfo:
    """Core information for describing field type and number"""

    # TODO: always put in id?
    def __init__(self, type_, number, description=None, id_=None):
        #: The type, one of INFO_TYPES or FORMAT_TYPES
        self.type = type_
        #: Number description, either an int or constant
        self.number = number
        #: Description for the header field, optional
        self.description = description
        #: The id of the field, optional.
        self.id = id_

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def __str__(self):
        return "FieldInfo({}, {}, {}, {})".format(
            *map(repr, [self.type, self.number, self.description, self.id])
        )

    def __repr__(self):
        return str(self)


# Reserved INFO keys ----------------------------------------------------------

#: Reserved fields for INFO from VCF v4.3
RESERVED_INFO = {
    # VCF v4.3, Section 1.6.1
    "AA": FieldInfo("String", 1, "Ancestral Allele"),
    "AC": FieldInfo(
        "Integer",
        "A",
        "Allele count in genotypes, for each ALT allele, in the " "same order as listed",
    ),
    "AD": FieldInfo("Integer", "R", "Total read depth for each allele"),
    "ADF": FieldInfo("Integer", "R", "Forward read depth for each allele"),
    "ADR": FieldInfo("Integer", "R", "Reverse read depth for each allele"),
    "AF": FieldInfo(
        "Float",
        "A",
        "Allele frequency for each ALT allele in the same order "
        "as listed: used for estimating from primary data not "
        "called genotypes",
    ),
    "AN": FieldInfo("Integer", 1, "Total number of alleles in called genotypes"),
    "BQ": FieldInfo("Float", 1, "RMS base quality at this position"),
    "CIGAR": FieldInfo(
        "String",
        "A",
        "CIGAR string describing how to align each ALT allele " "to the reference allele",
    ),
    "DB": FieldInfo("Flag", 0, "dbSNP membership"),
    "DP": FieldInfo(
        "Integer",
        1,
        "Combined depth across samples for small variants and "
        "Read Depth of segment containing breakend for SVs",
    ),
    "H2": FieldInfo("Flag", 0, "Membership in HapMap 2"),
    "H3": FieldInfo("Flag", 0, "Membership in HapMap 3"),
    "MQ": FieldInfo("Integer", 1, "RMS mapping quality"),
    "MQ0": FieldInfo("Integer", 1, "Number of MAPQ == 0 reads covering this record"),
    "NS": FieldInfo("Integer", 1, "Number of samples with data"),
    "SB": FieldInfo("Integer", 4, "Strand bias at this position"),
    "SOMATIC": FieldInfo(
        "Flag", 0, "Indicates that the record is a somatic mutation, " "for cancer genomics"
    ),
    "VALIDATED": FieldInfo("Flag", 0, "Validated by follow-up experiment"),
    "1000G": FieldInfo("Flag", 0, "Membership in 1000 Genomes"),
    # VCF v4.3, Section 3
    "IMPRECISE": FieldInfo("Flag", 0, "Imprecise structural variation"),
    "NOVEL": FieldInfo("Flag", 0, "Indicates a novel structural variation"),
    "END": FieldInfo(
        "Integer",
        1,
        "End position of the variant described in this record " "(for symbolic alleles)",
    ),
    "SVTYPE": FieldInfo("String", 1, "Type of structural variant"),
    "SVLEN": FieldInfo("Integer", 1, "Difference in length between REF and ALT alleles"),
    "CIPOS": FieldInfo("Integer", 2, "Confidence interval around POS for imprecise " "variants"),
    "CIEND": FieldInfo("Integer", 2, "Confidence interval around END for imprecise " "variants"),
    "HOMLEN": FieldInfo(
        "Integer", ".", "Length of base pair identical micro-homology at " "event breakpoints"
    ),
    "HOMSEQ": FieldInfo(
        "String", ".", "Sequence of base pair identical micro-homology at " "event breakpoints"
    ),
    "BKPTID": FieldInfo(
        "String", ".", "ID of the assembled alternate allele in the " "assembly file"
    ),
    "MEINFO": FieldInfo("String", 4, "Mobile element info of the form " "NAME,START,END,POLARITY"),
    "METRANS": FieldInfo(
        "String", 4, "Mobile element transduction info of the form " "CHR,START,END,POLARITY"
    ),
    "DGVID": FieldInfo("String", 1, "ID of this element in Database of Genomic Variation"),
    "DBVARID": FieldInfo("String", 1, "ID of this element in DBVAR"),
    "DBRIPID": FieldInfo("String", 1, "ID of this element in DBRIP"),
    "MATEID": FieldInfo("String", ".", "ID of mate breakends"),
    "PARID": FieldInfo("String", 1, "ID of partner breakend"),
    "EVENT": FieldInfo("String", 1, "ID of event associated to breakend"),
    "CILEN": FieldInfo(
        "Integer", 2, "Confidence interval around the inserted material " "between breakends"
    ),
    "DPADJ": FieldInfo("Integer", ".", "Read Depth of adjacency"),
    "CN": FieldInfo("Integer", 1, "Copy number of segment containing breakend"),
    "CNADJ": FieldInfo("Integer", ".", "Copy number of adjacency"),
    "CICN": FieldInfo("Integer", 2, "Confidence interval around copy number for the " "segment"),
    "CICNADJ": FieldInfo(
        "Integer", ".", "Confidence interval around copy number for the " "adjacency"
    ),
}

# Reserved FORMAT keys --------------------------------------------------------

RESERVED_FORMAT = {
    # VCF v 4.3, Section 1.6.2
    "AD": FieldInfo("Integer", "R", "Total, per-sample read depth"),
    "ADF": FieldInfo("Integer", "R", "Forward-strand, per-sample read depth"),
    "ADR": FieldInfo("Integer", "R", "Reverse-strand, per-sample read depth"),
    "DP": FieldInfo("Integer", 1, "Read depth at this position for this sample"),
    "EC": FieldInfo(
        "Integer", "A", "Expected alternate allele counts for each alternate " "allele"
    ),
    "FT": FieldInfo("String", "1", "Filters applied for this sample", "FORMAT/FT"),
    "GQ": FieldInfo("Integer", "G", "Phred-scale, conditional genotype quality"),
    "GP": FieldInfo("Float", "G", "Genotype posterior probabilities"),
    "GT": FieldInfo("String", 1, "Genotype call"),
    "GL": FieldInfo("Float", "G", "Log10-scaled likelihoods for genotypes"),
    "HQ": FieldInfo("Integer", 2, "Haplotype qualities"),
    "MQ": FieldInfo("Integer", 1, "RMS mapping quality"),
    "PL": FieldInfo("Integer", "G", "Phred-scaled genotype likelihoods, rounded to integers"),
    "PQ": FieldInfo("Integer", 1, "Phasing quality"),
    "PS": FieldInfo(
        "Integer",
        1,
        "Non-negative 32 bit integer giving phasing set " "for this sample and this chromosome",
    ),
    # VCF v4.3, Section 4
    "CN": FieldInfo("Integer", 1, "Copy number genotype for imprecise events"),
    "CNQ": FieldInfo("Float", 1, "Copy number genotype quality for imprecise events"),
    "CNL": FieldInfo("Float", "G", "Copy number genotype likelihood for imprecise events"),
    "CNP": FieldInfo("Float", "G", "Copy number posterior probabilities"),
    "NQ": FieldInfo("Integer", 1, "Phred style probability score that the variant is novel"),
    "HAP": FieldInfo("Integer", 1, "Unique haplotype identifier"),
    "AHAP": FieldInfo("Integer", 1, "Unique identifier of ancestral haplotype"),
}


# header files to enforce double-quoting for
QUOTE_FIELDS = ("Description", "Source", "Version")


def serialize_for_header(key, value):
    """Serialize value for the given mapping key for a VCF header line"""
    if key in QUOTE_FIELDS:
        return json.dumps(value)
    elif isinstance(value, str):
        if " " in value or "\t" in value:
            return json.dumps(value)
        else:
            return value
    elif isinstance(value, list):
        return "[{}]".format(", ".join(value))
    else:
        return str(value)


def header_without_lines(header, remove):
    """Return :py:class:`Header` without lines given in ``remove``

    ``remove`` is an iterable of pairs ``key``/``ID`` with the VCF header key
    and ``ID`` of entry to remove.  In the case that a line does not have
    a ``mapping`` entry, you can give the full value to remove.

    .. code-block:: python

        # header is a vcfpy.Header, e.g., as read earlier from file
        new_header = vcfpy.without_header_lines(
            header, [('assembly', None), ('FILTER', 'PASS')])
        # now, the header lines starting with "##assembly=" and the "PASS"
        # filter line will be missing from new_header
    """
    remove = set(remove)
    # Copy over lines that are not removed
    lines = []
    for line in header.lines:
        if hasattr(line, "mapping"):
            if (line.key, line.mapping.get("ID", None)) in remove:
                continue  # filter out
        else:
            if (line.key, line.value) in remove:
                continue  # filter out
        lines.append(line)
    return Header(lines, header.samples)


class Header:
    """Represent header of VCF file

    While this class allows mutating records, it should not be changed once it
    has been assigned to a writer.  Use :py:method:`~Header.copy` to create
    a copy that can be modified without problems.

    This class provides function for adding lines to a header and updating the
    supporting index data structures.  There is no explicit API for removing
    header lines, the best way is to reconstruct a new ``Header`` instance with
    a filtered list of header lines.
    """

    def __init__(self, lines=None, samples=None):
        #: ``list`` of :py:HeaderLine objects
        self.lines = lines or []
        #: :py:class:`SamplesInfo` object
        self.samples = samples
        # build indices for the different field types
        self._indices = self._build_indices()

    def _build_indices(self):
        """Build indices for the different field types"""
        result = {key: OrderedDict() for key in LINES_WITH_ID}
        for line in self.lines:
            if line.key in LINES_WITH_ID:
                result.setdefault(line.key, OrderedDict())
                if line.mapping["ID"] in result[line.key]:
                    warnings.warn(
                        ("Seen {} header more than once: {}, using first" "occurence").format(
                            line.key, line.mapping["ID"]
                        ),
                        DuplicateHeaderLineWarning,
                    )
                else:
                    result[line.key][line.mapping["ID"]] = line
            else:
                result.setdefault(line.key, [])
                result[line.key].append(line)
        return result

    def copy(self):
        """Return a copy of this header"""
        return Header([line.copy() for line in self.lines], self.samples.copy())

    def add_filter_line(self, mapping):
        """Add FILTER header line constructed from the given mapping

        :param mapping: ``OrderedDict`` with mapping to add.  It is
            recommended to use ``OrderedDict`` over ``dict`` as this makes
            the result reproducible
        :return: ``False`` on conflicting line and ``True`` otherwise
        """
        return self.add_line(FilterHeaderLine.from_mapping(mapping))

    def add_contig_line(self, mapping):
        """Add "contig" header line constructed from the given mapping

        :param mapping: ``OrderedDict`` with mapping to add.  It is
            recommended to use ``OrderedDict`` over ``dict`` as this makes
            the result reproducible
        :return: ``False`` on conflicting line and ``True`` otherwise
        """
        return self.add_line(ContigHeaderLine.from_mapping(mapping))

    def add_info_line(self, mapping):
        """Add INFO header line constructed from the given mapping

        :param mapping: ``OrderedDict`` with mapping to add.  It is
            recommended to use ``OrderedDict`` over ``dict`` as this makes
            the result reproducible
        :return: ``False`` on conflicting line and ``True`` otherwise
        """
        return self.add_line(InfoHeaderLine.from_mapping(mapping))

    def add_format_line(self, mapping):
        """Add FORMAT header line constructed from the given mapping

        :param mapping: ``OrderedDict`` with mapping to add.  It is
            recommended to use ``OrderedDict`` over ``dict`` as this makes
            the result reproducible
        :return: ``False`` on conflicting line and ``True`` otherwise
        """
        return self.add_line(FormatHeaderLine.from_mapping(mapping))

    def format_ids(self):
        """Return list of all format IDs"""
        return list(self._indices["FORMAT"].keys())

    def filter_ids(self):
        """Return list of all filter IDs"""
        return list(self._indices["FILTER"].keys())

    def info_ids(self):
        """Return list of all info IDs"""
        return list(self._indices["INFO"].keys())

    def get_lines(self, key):
        """Return header lines having the given ``key`` as their type"""
        if key in self._indices:
            return self._indices[key].values()
        else:
            return []

    def has_header_line(self, key, id_):
        """Return whether there is a header line with the given ID of the
        type given by ``key``

        :param key: The VCF header key/line type.
        :param id_: The ID value to compare fore

        :return: ``True`` if there is a header line starting with ``##${key}=``
            in the VCF file having the mapping entry ``ID`` set to ``id_``.
        """
        if key not in self._indices:
            return False
        else:
            return id_ in self._indices[key]

    def add_line(self, header_line):
        """Add header line, updating any necessary support indices

        :return: ``False`` on conflicting line and ``True`` otherwise
        """
        self.lines.append(header_line)
        self._indices.setdefault(header_line.key, OrderedDict())
        if not hasattr(header_line, "mapping"):
            return False  # no registration required
        if self.has_header_line(header_line.key, header_line.mapping["ID"]):
            warnings.warn(
                (
                    "Detected duplicate header line with type {} and ID {}. "
                    "Ignoring this and subsequent one"
                ).format(header_line.key, header_line.mapping["ID"]),
                DuplicateHeaderLineWarning,
            )
            return False
        else:
            self._indices[header_line.key][header_line.mapping["ID"]] = header_line
            return True

    def get_info_field_info(self, key):
        """Return :py:class:`FieldInfo` for the given INFO field"""
        return self._get_field_info("INFO", key)

    def get_format_field_info(self, key):
        """Return :py:class:`FieldInfo` for the given INFO field"""
        return self._get_field_info("FORMAT", key)

    def _get_field_info(self, type_, key):
        result = self._indices[type_].get(key)
        if result:
            return result
        if key in RESERVED_INFO:
            res = FieldInfo(RESERVED_INFO[key].type, RESERVED_INFO[key].number)
        else:
            res = FieldInfo("String", HEADER_NUMBER_UNBOUNDED)
        warnings.warn(
            "{} {} not found using {}/{} instead".format(type_, key, res.type, repr(res.number)),
            FieldInfoNotFound,
        )
        return res

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.lines, self.samples) == (other.lines, other.samples)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return (self.lines, self.samples) != (other.lines, other.samples)
        return NotImplemented

    def __hash__(self):
        raise TypeError("Unhashable type: Header")

    def __str__(self):
        tpl = "Header(lines={}, samples={})"
        return tpl.format(*map(repr, (self.lines, self.samples)))

    def __repr__(self):
        return str(self)


class HeaderLine:
    """Base class for VCF header lines"""

    def __init__(self, key, value):
        #: ``str`` with key of header line
        self.key = key
        # ``str`` with raw value of header line
        self._value = value

    def copy(self):
        """Return a copy"""
        return self.__class__(self.key, self.value)

    @property
    def value(self):
        return self._value

    def serialize(self):
        """Return VCF-serialized version of this header line"""
        return "".join(("##", self.key, "=", self.value))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.key, self.value) == (other.key, other.value)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return (self.key, self.value) != (other.key, other.value)
        return NotImplemented

    def __hash__(self):
        raise TypeError("Unhashable type: HeaderLine")

    def __str__(self):
        return "HeaderLine({}, {})".format(*map(repr, (self.key, self.value)))

    def __repr__(self):
        return str(self)


def mapping_to_str(mapping):
    """Convert mapping to string"""
    result = ["<"]
    for i, (key, value) in enumerate(mapping.items()):
        if i > 0:
            result.append(",")
        result += [key, "=", serialize_for_header(key, value)]
    result += [">"]
    return "".join(result)


class SimpleHeaderLine(HeaderLine):
    """Base class for simple header lines, currently contig and filter
    header lines

    Don't use this class directly but rather the sub classes.

    :raises: :py:class:`vcfpy.exceptions.InvalidHeaderException` in
        the case of missing key ``"ID"``
    """

    def __init__(self, key, value, mapping):
        super().__init__(key, value)
        # check existence of key "ID"
        if "ID" not in mapping:
            raise exceptions.InvalidHeaderException(
                'Missing key "ID" in header line "{}={}"'.format(key, value)
            )
        #: ``collections.OrderedDict`` with key/value mapping of the attributes
        self.mapping = OrderedDict(mapping.items())

    def copy(self):
        """Return a copy"""
        mapping = OrderedDict(self.mapping.items())
        return self.__class__(self.key, self.value, mapping)

    @property
    def value(self):
        return mapping_to_str(self.mapping)

    def serialize(self):
        return "".join(map(str, ["##", self.key, "=", self.value]))

    def __str__(self):
        return "SimpleHeaderLine({}, {}, {})".format(
            *map(repr, (self.key, self.value, self.mapping))
        )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.key, self.value, self.mapping) == (other.key, other.value, other.mapping)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return (self.key, self.value, self.mapping) != (other.key, other.value, other.mapping)
        return NotImplemented


class AltAlleleHeaderLine(SimpleHeaderLine):
    """Alternative allele header line

    Mostly used for defining symbolic alleles for structural variants and
    IUPAC ambiguity codes
    """

    @classmethod
    def from_mapping(klass, mapping):
        """Construct from mapping, not requiring the string value"""
        return AltAlleleHeaderLine("ALT", mapping_to_str(mapping), mapping)

    def __init__(self, key, value, mapping):
        super().__init__(key, value, mapping)
        #: name of the alternative allele
        self.id = self.mapping["ID"]

    def __hash__(self):
        raise TypeError("Unhashable type: AltAlleleHeaderLine")

    def __str__(self):
        return "AltAlleleHeaderLine({}, {}, {})".format(
            *map(repr, (self.key, self.value, self.mapping))
        )


class ContigHeaderLine(SimpleHeaderLine):
    """Contig header line

    Most importantly, parses the ``'length'`` key into an integer
    """

    @classmethod
    def from_mapping(klass, mapping):
        """Construct from mapping, not requiring the string value"""
        return ContigHeaderLine("contig", mapping_to_str(mapping), mapping)

    def __init__(self, key, value, mapping):
        super().__init__(key, value, mapping)
        # convert 'length' entry to integer if possible
        if "length" in self.mapping:
            mapping["length"] = int(mapping["length"])
        else:
            warnings.warn(
                'Field "length" not found in header line {}={}'.format(key, value),
                FieldInfoNotFound,
            )
        #: name of the contig
        self.id = self.mapping["ID"]
        #: length of the contig, ``None`` if missing
        self.length = self.mapping.get("length")

    def __hash__(self):
        raise TypeError("Unhashable type: ContigHeaderLine")

    def __str__(self):
        return "ContigHeaderLine({}, {}, {})".format(
            *map(repr, (self.key, self.value, self.mapping))
        )


class FilterHeaderLine(SimpleHeaderLine):
    """FILTER header line"""

    @classmethod
    def from_mapping(klass, mapping):
        """Construct from mapping, not requiring the string value"""
        return FilterHeaderLine("FILTER", mapping_to_str(mapping), mapping)

    def __init__(self, key, value, mapping):
        super().__init__(key, value, mapping)
        # check for "Description" key
        if "Description" not in self.mapping:
            warnings.warn(
                'Field "Description" not found in header line {}={}'.format(key, value),
                FieldInfoNotFound,
            )
        #: token for the filter
        self.id = self.mapping["ID"]
        #: description for the filter, ``None`` if missing
        self.description = self.mapping.get("Description")

    def __hash__(self):
        raise TypeError("Unhashable type: FilterHeaderLine")

    def __str__(self):
        return "FilterHeaderLine({}, {}, {})".format(
            *map(repr, (self.key, self.value, self.mapping))
        )


class MetaHeaderLine(SimpleHeaderLine):
    """Alternative allele header line

    Used for defining set of valid values for samples keys
    """

    @classmethod
    def from_mapping(klass, mapping):
        """Construct from mapping, not requiring the string value"""
        return MetaHeaderLine("META", mapping_to_str(mapping), mapping)

    def __init__(self, key, value, mapping):
        super().__init__(key, value, mapping)
        #: name of the alternative allele
        self.id = self.mapping["ID"]

    def __hash__(self):
        raise TypeError("Unhashable type: MetaHeaderLine")

    def __str__(self):
        return "MetaHeaderLine({}, {}, {})".format(*map(repr, (self.key, self.value, self.mapping)))


class PedigreeHeaderLine(SimpleHeaderLine):
    """Header line for defining a pedigree entry"""

    @classmethod
    def from_mapping(klass, mapping):
        """Construct from mapping, not requiring the string value"""
        return PedigreeHeaderLine("PEDIGREE", mapping_to_str(mapping), mapping)

    def __init__(self, key, value, mapping):
        super().__init__(key, value, mapping)
        #: name of the alternative allele
        self.id = self.mapping["ID"]

    def __hash__(self):
        raise TypeError("Unhashable type: PedigreeHeaderLine")

    def __str__(self):
        return "PedigreeHeaderLine({}, {}, {})".format(
            *map(repr, (self.key, self.value, self.mapping))
        )


class SampleHeaderLine(SimpleHeaderLine):
    """Header line for defining a SAMPLE entry"""

    @classmethod
    def from_mapping(klass, mapping):
        """Construct from mapping, not requiring the string value"""
        return SampleHeaderLine("SAMPLE", mapping_to_str(mapping), mapping)

    def __init__(self, key, value, mapping):
        super().__init__(key, value, mapping)
        #: name of the alternative allele
        self.id = self.mapping["ID"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.key, self.value, self.mapping) == (other.key, other.value, other.mapping)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return (self.key, self.value, self.mapping) != (other.key, other.value, other.mapping)
        return NotImplemented

    def __hash__(self):
        raise TypeError("Unhashable type: SampleHeaderLine")

    def __str__(self):
        return "SampleHeaderLine({}, {}, {})".format(
            *map(repr, (self.key, self.value, self.mapping))
        )


class CompoundHeaderLine(HeaderLine):
    """Base class for compound header lines, currently format and header lines

    Compound header lines describe fields that can have more than one entry.

    Don't use this class directly but rather the sub classes.
    """

    def __init__(self, key, value, mapping):
        super().__init__(key, value)
        #: OrderedDict with key/value mapping
        self.mapping = OrderedDict(mapping.items())
        # check that 'Number' is given and use "." otherwise
        if "Number" not in self.mapping:
            warnings.warn(
                '[vcfpy] WARNING: missing number, using unbounded/"." instead', FieldMissingNumber
            )
            self.mapping["Number"] = "."
        try:
            self.mapping["Number"] = self._parse_number(self.mapping["Number"])
        except ValueError:
            warnings.warn(
                ("[vcfpy] WARNING: invalid number {}, using " 'unbounded/"." instead').format(
                    self.mapping["Number"]
                ),
                FieldInvalidNumber,
            )
            self.mapping["Number"] = "."

    def copy(self):
        """Return a copy"""
        mapping = OrderedDict(self.mapping.items())
        return self.__class__(self.key, self.value, mapping)

    @classmethod
    def _parse_number(klass, number):
        """Parse ``number`` into an ``int`` or return ``number`` if a valid
        expression for a INFO/FORMAT "Number".

        :param str number: ``str`` to parse and check
        """
        try:
            return int(number)
        except ValueError as e:
            if number in VALID_NUMBERS:
                return number
            else:
                raise e

    @property
    def value(self):
        return mapping_to_str(self.mapping)

    def serialize(self):
        return "".join(map(str, ["##", self.key, "=", self.value]))

    def __str__(self):
        return "CompoundHeaderLine({}, {}, {})".format(
            *map(repr, (self.key, self.value, self.mapping))
        )


class InfoHeaderLine(CompoundHeaderLine):
    """Header line for INFO fields

    Note that the ``Number`` field will be parsed into an ``int`` if
    possible.  Otherwise, the constants ``HEADER_NUMBER_*`` will be used.
    """

    @classmethod
    def from_mapping(klass, mapping):
        """Construct from mapping, not requiring the string value"""
        return InfoHeaderLine("INFO", mapping_to_str(mapping), mapping)

    def __init__(self, key, value, mapping):
        super().__init__(key, value, mapping)
        #: key in the INFO field
        self.id = self.mapping["ID"]
        # check for "Number" field
        self.number = self.mapping["Number"]
        # check for "Type" field
        type_ = self.mapping.get("Type")
        if "Type" not in self.mapping:
            warnings.warn(
                ('Field "Type" not found in header line, using String ' "instead {}={}").format(
                    key, value
                ),
                HeaderInvalidType,
            )
            type_ = "String"
        if "Type" in self.mapping and type_ not in INFO_TYPES:
            warnings.warn(
                (
                    "Invalid INFO value type {} in header line, using String " "instead, {}={}"
                ).format(self.mapping["Type"], key, value),
                HeaderInvalidType,
            )
            type_ = "String"
        #: value type
        self.type = type_
        # check for "Description" key
        if "Description" not in self.mapping:
            warnings.warn(
                'Field "Description" not found in header line {}={}'.format(key, value),
                HeaderMissingDescription,
            )
        #: description, should be given, ``None`` if not given
        self.description = self.mapping.get("Description")
        #: source of INFO field, ``None`` if not given
        self.source = self.mapping.get("Source")
        #: version of INFO field, ``None`` if not given
        self.version = self.mapping.get("Version")

    def __hash__(self):
        raise TypeError("Unhashable type: InfoHeaderLine")

    def __str__(self):
        return "InfoHeaderLine({}, {}, {})".format(*map(repr, (self.key, self.value, self.mapping)))


class FormatHeaderLine(CompoundHeaderLine):
    """Header line for FORMAT fields"""

    @classmethod
    def from_mapping(klass, mapping):
        """Construct from mapping, not requiring the string value"""
        return FormatHeaderLine("FORMAT", mapping_to_str(mapping), mapping)

    def __init__(self, key, value, mapping):
        super().__init__(key, value, mapping)
        #: key in the INFO field
        self.id = self.mapping["ID"]
        # check for "Number" field
        self.number = self.mapping["Number"]
        # check for "Type" field
        type_ = self.mapping.get("Type")
        if "Type" not in self.mapping:
            warnings.warn(
                ('Field "Type" not found in header line, using String ' "instead {}={}").format(
                    key, value
                ),
                HeaderInvalidType,
            )
            type_ = "String"
        if "Type" in self.mapping and type_ not in FORMAT_TYPES:
            warnings.warn(
                (
                    "Invalid FORMAT value type {} in header line, using String " "instead, {}={}"
                ).format(self.mapping["Type"], key, value),
                HeaderInvalidType,
            )
            type_ = "String"
        #: value type
        self.type = type_
        # check for "Description" key
        if "Description" not in self.mapping:
            warnings.warn(
                'Field "Description" not found in header line {}={}'.format(key, value),
                HeaderMissingDescription,
            )
        #: description, should be given, ``None`` if not given
        self.description = self.mapping.get("Description")
        #: source of INFO field, ``None`` if not given
        self.source = self.mapping.get("Source")
        #: version of INFO field, ``None`` if not given
        self.version = self.mapping.get("Version")

    def __hash__(self):
        raise TypeError("Unhashable type: FormatHeaderLine")

    def __str__(self):
        return "FormatHeaderLine({}, {}, {})".format(
            *map(repr, (self.key, self.value, self.mapping))
        )


class SamplesInfos:
    """Helper class for handling the samples in VCF files

    The purpose of this class is to decouple the sample name list somewhat
    from :py:class:`Header`.  This encapsulates subsetting samples for which
    the genotype should be parsed and reordering samples into output files.

    Note that when subsetting is used and the records are to be written out
    again then the ``FORMAT`` field must not be touched.
    """

    def __init__(self, sample_names, parsed_samples=None):
        #: list of sample that are read from/written to the VCF file at
        #: hand in the given order
        self.names = list(sample_names)
        #: ``set`` with the samples for which the genotype call fields should
        #: be read; can be used for partial parsing (speedup) and defaults
        #: to the full list of samples, None if all are parsed
        self.parsed_samples = parsed_samples
        if self.parsed_samples:
            self.parsed_samples = set(self.parsed_samples)
            assert self.parsed_samples <= set(self.names), "Must be subset!"
        #: mapping from sample name to index
        self.name_to_idx = dict([(name, idx) for idx, name in enumerate(self.names)])

    def copy(self):
        """Return a copy of the object"""
        return SamplesInfos(self.names)

    def is_parsed(self, name):
        """Return whether the sample name is parsed"""
        return (not self.parsed_samples) or name in self.parsed_samples

    def __hash__(self):
        raise TypeError("Unhashable type: SamplesInfos")

    def __str__(self):
        tpl = "SamplesInfos(names={}, name_to_idx={})"
        return tpl.format(self.names, pprint.pformat(self.name_to_idx, width=10**10))

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.names == other.names
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return self.names != other.names
        return NotImplemented



# -*- coding: utf-8 -*-
"""Reading of VCF header from plain and bgzip-ed file
"""

import os
import sys

from vcfpy import reader

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


def test_read_text():
    path = os.path.join(os.path.dirname(__file__), "vcfs/from_vcf43.vcf")
    r = reader.Reader.from_path(path)
    assert r.parser
    assert r.header
    assert len(r.header.lines) == 18
    EXPECTED = "HeaderLine('fileformat', 'VCFv4.3')"
    assert str(r.header.lines[0]) == EXPECTED
    if sys.version_info < (3, 6):
        EXPECTED = (
            "FormatHeaderLine('FORMAT', '<ID=HQ,Number=2,Type=Integer,Description=\"Haplotype Quality\">', "
            "OrderedDict([('ID', 'HQ'), ('Number', 2), ('Type', 'Integer'), ('Description', 'Haplotype Quality')]))"
        )
    else:
        EXPECTED = (
            "FormatHeaderLine('FORMAT', '<ID=HQ,Number=2,Type=Integer,Description=\"Haplotype Quality\">', "
            "{'ID': 'HQ', 'Number': 2, 'Type': 'Integer', 'Description': 'Haplotype Quality'})"
        )
    assert str(r.header.lines[-1]) == EXPECTED
    assert r.header.samples
    assert r.header.samples.names == ["NA00001", "NA00002", "NA00003"]


def test_read_bgzip():
    path = os.path.join(os.path.dirname(__file__), "vcfs/from_vcf43.vcf.gz")
    r = reader.Reader.from_path(path)
    assert r.parser
    assert r.header
    assert len(r.header.lines) == 18
    EXPECTED = "HeaderLine('fileformat', 'VCFv4.3')"
    assert str(r.header.lines[0]) == EXPECTED
    if sys.version_info < (3, 6):
        EXPECTED = (
            "FormatHeaderLine('FORMAT', '<ID=HQ,Number=2,Type=Integer,Description=\"Haplotype Quality\">', "
            "OrderedDict([('ID', 'HQ'), ('Number', 2), ('Type', 'Integer'), ('Description', 'Haplotype Quality')]))"
        )
    else:
        EXPECTED = (
            "FormatHeaderLine('FORMAT', '<ID=HQ,Number=2,Type=Integer,Description=\"Haplotype Quality\">', "
            "{'ID': 'HQ', 'Number': 2, 'Type': 'Integer', 'Description': 'Haplotype Quality'})"
        )
    assert str(r.header.lines[-1]) == EXPECTED
    assert r.header.samples
    assert r.header.samples.names == ["NA00001", "NA00002", "NA00003"]


def test_read_text_no_samples():
    path = os.path.join(os.path.dirname(__file__), "vcfs/from_vcf43_nosamples.vcf")
    r = reader.Reader.from_path(path)
    assert r.parser
    assert r.header
    assert len(r.header.lines) == 18
    EXPECTED = "HeaderLine('fileformat', 'VCFv4.3')"
    assert str(r.header.lines[0]) == EXPECTED
    if sys.version_info < (3, 6):
        EXPECTED = (
            "FormatHeaderLine('FORMAT', '<ID=HQ,Number=2,Type=Integer,Description=\"Haplotype Quality\">', "
            "OrderedDict([('ID', 'HQ'), ('Number', 2), ('Type', 'Integer'), ('Description', 'Haplotype Quality')]))"
        )
    else:
        EXPECTED = (
            "FormatHeaderLine('FORMAT', '<ID=HQ,Number=2,Type=Integer,Description=\"Haplotype Quality\">', "
            "{'ID': 'HQ', 'Number': 2, 'Type': 'Integer', 'Description': 'Haplotype Quality'})"
        )
    assert str(r.header.lines[-1]) == EXPECTED
    assert r.header.samples
    assert r.header.samples.names == []



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from itertools import chain
import os.path
import sys

from setuptools import setup

import versioneer


def parse_requirements(path):
    """Parse ``requirements.txt`` at ``path``."""
    requirements = []
    with open(path, "rt") as reqs_f:
        for line in reqs_f:
            line = line.strip()
            if line.startswith("-r"):
                fname = line.split()[1]
                inner_path = os.path.join(os.path.dirname(path), fname)
                requirements += parse_requirements(inner_path)
            elif line != "" and not line.startswith("#"):
                requirements.append(line)
    return requirements


with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

base_reqs = parse_requirements("requirements.txt")

# Add cyordereddict for Python <=3.5 for performance boost
if sys.version_info[:2] < (3, 6):
    pre36_reqs = parse_requirements("requirements/pre36.txt")
else:
    pre36_reqs = []

requirements = chain(base_reqs, pre36_reqs)

test_requirements = parse_requirements("requirements/test.txt")

setup(
    name="vcfpy",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description=("Python 3 VCF library with good support for both reading and writing"),
    long_description=readme + "\n\n" + history,
    author="Manuel Holtgrewe",
    author_email="manuel.holtgrewe@bihealth.de",
    url="https://github.com/bihealth/vcfpy",
    packages=["vcfpy"],
    package_dir={"vcfpy": "vcfpy"},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords="vcfpy",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    test_suite="tests",
    tests_require=test_requirements,
)



# -*- coding: utf-8 -*-
"""Test for the helper routines in the vcfpy.parser module"""

import pytest

from vcfpy import parser

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


# parser.split_quoted_string() ------------------------------------------------


def test_split_quoted_string_one_noquote():
    INPUT = "foo=bar"
    EXPECTED = ["foo=bar"]
    assert EXPECTED == parser.split_quoted_string(INPUT)


def test_split_quoted_string_two_noquote():
    INPUT = "foo=bar,bar=baz"
    EXPECTED = ["foo=bar", "bar=baz"]
    assert EXPECTED == parser.split_quoted_string(INPUT)


def test_split_quoted_string_one_quote():
    INPUT = 'foo="bar,,"asdf'
    EXPECTED = ['foo="bar,,"asdf']
    assert EXPECTED == parser.split_quoted_string(INPUT)


def test_split_quoted_string_two_quote():
    INPUT = 'foo="bar,,"asdf,"bar"=baz'
    EXPECTED = ['foo="bar,,"asdf', '"bar"=baz']
    assert EXPECTED == parser.split_quoted_string(INPUT)


def test_split_quoted_string_one_escape():
    INPUT = 'foo="bar\\"asdf\\"asdf"'
    EXPECTED = ['foo="bar\\"asdf\\"asdf"']
    assert EXPECTED == parser.split_quoted_string(INPUT)


def test_split_quoted_string_two_escape():
    INPUT = 'foo="bar\\",,asdf","bar"=baz'
    EXPECTED = ['foo="bar\\",,asdf"', '"bar"=baz']
    assert EXPECTED == parser.split_quoted_string(INPUT)


def test_split_quoted_string_array_syntax_simple():
    INPUT = "foo=[1, 2, 3],bar=[baz, gnaa]"
    EXPECTED = ["foo=[1, 2, 3]", "bar=[baz, gnaa]"]
    assert EXPECTED == parser.split_quoted_string(INPUT)


def test_split_quoted_string_array_syntax_recursion():
    """Support for only one level, failure is OK here"""
    INPUT = "foo=[1, 2, 3],bar=[baz, [], gnaa]"
    EXPECTED = ["foo=[1, 2, 3]", "bar=[baz, []", " gnaa]"]
    assert EXPECTED == parser.split_quoted_string(INPUT)


# parser.VCFheaderLineParser.parse_mapping() ----------------------------------


def test_vcf_header_line_parser_parse_mapping_simple():
    INPUT = r"<key=value,key2=value2>"
    EXPECTED = (("key", "value"), ("key2", "value2"))
    assert EXPECTED == tuple(parser.parse_mapping(INPUT).items())


def test_vcf_header_line_parser_parse_mapping_flag():
    INPUT = r"<key=value,key2=value,yay>"
    EXPECTED = (("key", "value"), ("key2", "value"), ("yay", True))
    parser.MappingHeaderLineParser(None)
    assert EXPECTED == tuple(parser.parse_mapping(INPUT).items())


def test_vcf_header_line_parser_parse_mapping_quoted():
    INPUT = r'<key=value,key2="value,value">'
    EXPECTED = (("key", "value"), ("key2", "value,value"))
    parser.MappingHeaderLineParser(None)
    assert EXPECTED == tuple(parser.parse_mapping(INPUT).items())


def test_vcf_header_line_parser_parse_mapping_escaped():
    INPUT = r'<key=value,key2="value,value=\"asdf">'
    EXPECTED = (("key", "value"), ("key2", 'value,value="asdf'))
    parser.MappingHeaderLineParser(None)
    assert EXPECTED == tuple(parser.parse_mapping(INPUT).items())



# -*- coding: utf-8 -*-
"""Parsing of VCF files from ``file``-like objects
"""

import gzip
import os

import pysam

from . import parser

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


class Reader:
    """Class for parsing of files from ``file``-like objects

    Instead of using the constructor, use the class methods
    :py:meth:`~Reader.from_stream` and
    :py:meth:`~Reader.from_path`.

    On construction, the header will be read from the file which can cause
    problems.  After construction, :py:class:`~Reader` can be used as
    an iterable of :py:class:`~vcfpy.record.Record`.

    :raises: :py:class:`~vcfpy.exceptions.InvalidHeaderException` in the case
        of problems reading the header

    .. note::
        It is important to note that the ``header`` member is used during
        the parsing of the file.  **If you need a modified version then
        create a copy, e.g., using :py:method:`~vcfpy.header.Header.copy`**.

    .. note::
        If you use the ``parsed_samples`` feature and you write out
        records then you must not change the ``FORMAT`` of the record.
    """

    @classmethod
    def from_stream(
        klass, stream, path=None, tabix_path=None, record_checks=None, parsed_samples=None
    ):
        """Create new :py:class:`Reader` from file

        .. note::
            If you use the ``parsed_samples`` feature and you write out
            records then you must not change the ``FORMAT`` of the record.

        :param stream: ``file``-like object to read from
        :param path: optional string with path to store (for display only)
        :param list record_checks: record checks to perform, can contain
            'INFO' and 'FORMAT'
        :param list parsed_samples: ``list`` of ``str`` values with names of
            samples to parse call information for (for speedup); leave to
            ``None`` for ignoring
        """
        record_checks = record_checks or []
        if tabix_path and not path:
            raise ValueError("Must give path if tabix_path is given")
        return Reader(
            stream=stream,
            path=path,
            tabix_path=tabix_path,
            record_checks=record_checks,
            parsed_samples=parsed_samples,
        )

    @classmethod
    def from_path(klass, path, tabix_path=None, record_checks=None, parsed_samples=None):
        """Create new :py:class:`Reader` from path

        .. note::
            If you use the ``parsed_samples`` feature and you write out
            records then you must not change the ``FORMAT`` of the record.

        :param path: the path to load from (converted to ``str`` for
            compatibility with ``path.py``)
        :param tabix_path: optional string with path to TBI index,
            automatic inferral from ``path`` will be tried on the fly
            if not given
        :param list record_checks: record checks to perform, can contain
            'INFO' and 'FORMAT'
        """
        record_checks = record_checks or []
        path = str(path)
        if path.endswith(".gz"):
            f = gzip.open(path, "rt")
            if not tabix_path:
                tabix_path = path + ".tbi"
                if not os.path.exists(tabix_path):
                    tabix_path = None  # guessing path failed
        else:
            f = open(path, "rt")
        return klass.from_stream(
            stream=f,
            path=path,
            tabix_path=tabix_path,
            record_checks=record_checks,
            parsed_samples=parsed_samples,
        )

    def __init__(self, stream, path=None, tabix_path=None, record_checks=None, parsed_samples=None):
        #: stream (``file``-like object) to read from
        self.stream = stream
        #: optional ``str`` with the path to the stream
        self.path = path
        #: optional ``str`` with path to tabix file
        self.tabix_path = tabix_path
        #: checks to perform on records, can contain 'FORMAT' and 'INFO'
        self.record_checks = tuple(record_checks or [])
        #: if set, list of samples to parse for
        self.parsed_samples = parsed_samples
        #: the ``pysam.TabixFile`` used for reading from index bgzip-ed VCF;
        #: constructed on the fly
        self.tabix_file = None
        # the iterator through the Tabix file to use
        self.tabix_iter = None
        #: the parser to use
        self.parser = parser.Parser(stream, self.path, self.record_checks)
        #: the Header
        self.header = self.parser.parse_header(parsed_samples)

    def fetch(self, chrom_or_region, begin=None, end=None):
        """Jump to the start position of the given chromosomal position
        and limit iteration to the end position

        :param str chrom_or_region: name of the chromosome to jump to if
            begin and end are given and a samtools region string otherwise
            (e.g. "chr1:123,456-123,900").
        :param int begin: 0-based begin position (inclusive)
        :param int end: 0-based end position (exclusive)
        """
        if begin is not None and end is None:
            raise ValueError("begin and end must both be None or neither")
        # close tabix file if any and is open
        if self.tabix_file and not self.tabix_file.closed:
            self.tabix_file.close()
        # open tabix file if not yet open
        if not self.tabix_file or self.tabix_file.closed:
            self.tabix_file = pysam.TabixFile(filename=self.path, index=self.tabix_path)
        # jump to the next position
        if begin is None:
            self.tabix_iter = self.tabix_file.fetch(region=chrom_or_region)
        else:
            self.tabix_iter = self.tabix_file.fetch(reference=chrom_or_region, start=begin, end=end)
        return self

    def close(self):
        """Close underlying stream"""
        if self.tabix_file and not self.tabix_file.closed:
            self.tabix_file.close()
        if self.stream:
            self.stream.close()

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        self.close()

    def __iter__(self):
        return self

    def __next__(self):
        """Return next object from file

        :returns:
        :raises: ``vcfpy.exceptions.InvalidRecordException`` in the case of
            problems reading the record
        :raises: ``StopException`` if at end
        """
        if self.tabix_iter:
            return self.parser.parse_line(str(next(self.tabix_iter)))
        else:
            result = self.parser.parse_next_record()
            if result is None:
                raise StopIteration()
            else:
                return result
